import pytest
import tenacity

from osa.configs import options


def test_analysis_stage(running_analysis_dir):
    from osa.workflow.stages import AnalysisStage

    options.simulate = False
    options.directory = running_analysis_dir

    # First step of the analysis
    r0_file = "input_r0.fits"
    pedestal_file = "drs4_file.h5"
    calibration_file = "calib_file.h5"
    drive_file = "drive_log.txt"

    cmd = [
        "lstchain_data_r0_to_dl1",
        f"--input-file={r0_file}",
        f"--pedestal-file={pedestal_file}",
        f"--calibration-file={calibration_file}",
        f"--drive-file={drive_file}",
    ]
    stage = AnalysisStage(run="01000.0001", command_args=cmd)
    assert stage.rc is None
    assert stage.show_command() == " ".join(cmd)
    with pytest.raises(tenacity.RetryError):
        stage.execute()
    assert stage.rc == 2
    # Check that the stage is marked as failed in the history file
    with open(stage.history_file, "r") as f:
        lines = f.readlines()
        assert len(lines) >= 3
    # Check that the last element in the last line is the rc 2
    assert lines[-1].split(" ")[0] == stage.run
    assert lines[-1].split(" ")[1] == cmd[0]
    assert lines[-1].split(" ")[-1] == "2\n"

    # Second step
    cmd = [
        "lstchain_dl1ab",
        "--input-file=dl1a_file.h5",
        "--output-file=dl1b_file.h5",
    ]
    stage = AnalysisStage(run="01000.0001", command_args=cmd)
    assert stage.rc is None
    assert stage.show_command() == " ".join(cmd)
    with pytest.raises(tenacity.RetryError):
        stage.execute()
    assert stage.rc == 1
    # Check that the stage is marked as failed in the history file
    with open(stage.history_file, "r") as f:
        lines = f.readlines()
        assert len(lines) >= 6
    # Check that the last element in the last line is the step rc
    assert lines[-1].split(" ")[0] == stage.run
    assert lines[-1].split(" ")[1] == cmd[0]
    assert lines[-1].split(" ")[-1] == "1\n"

    # Third step
    cmd = ["lstchain_check_dl1", "--input-file=dl1_file.h5", "--batch"]
    stage = AnalysisStage(run="01000.0001", command_args=cmd)
    assert stage.rc is None
    assert stage.show_command() == " ".join(cmd)
    with pytest.raises(tenacity.RetryError):
        stage.execute()
    assert stage.rc == 255
    # Check that the stage is marked as failed in the history file
    with open(stage.history_file, "r") as f:
        lines = f.readlines()
        assert len(lines) >= 9
    # Check that the last element in the last line is the step rc
    assert lines[-1].split(" ")[0] == stage.run
    assert lines[-1].split(" ")[1] == cmd[0]
    assert lines[-1].split(" ")[-1] == "255\n"


def test_calibration_steps(running_analysis_dir):
    from osa.workflow.stages import DRS4PedestalStage, ChargeCalibrationStage
    from osa.scripts.calibration_pipeline import drs4_pedestal_command, calibration_file_command

    options.simulate = False
    options.directory = running_analysis_dir

    cmd1 = drs4_pedestal_command(998)
    cmd2 = calibration_file_command(drs4_pedestal_run_id=998, pedcal_run_id=999)
    step1 = DRS4PedestalStage(run="00998", run_pedcal="00999", command_args=cmd1)
    step2 = ChargeCalibrationStage(run="00999", command_args=cmd2)

    with pytest.raises(tenacity.RetryError):
        step1.execute()

    with pytest.raises(tenacity.RetryError):
        step2.execute()

    assert step1.history_file == step2.history_file
    # Check that the error is in the history file
    with open(step1.history_file, "r") as f:
        lines = f.readlines()
        assert len(lines) == 6
    # Check that the last element in the last line is the rc
    assert lines[0].split(" ")[0] == step1.run
    assert lines[-1].split(" ")[0] == step2.run
    assert lines[0].split(" ")[1] == cmd1[0]
    assert lines[-1].split(" ")[1] == cmd2[0]
    assert lines[0].split(" ")[-1] == "1\n"
    assert lines[-1].split(" ")[-1] == "1\n"



#!/usr/bin/env python

"""
Orchestrator script that creates and execute the calibration sequence and
prepares a SLURM job array which launches the data sequences for every subrun.
"""

import logging
import os
from decimal import Decimal

from osa import osadb
from osa.configs import options
from osa.configs.config import cfg
from osa.job import (
    set_queue_values,
    prepare_jobs,
    submit_jobs,
    get_sacct_output,
    get_squeue_output,
    run_sacct,
    run_squeue,
)
from osa.nightsummary.extract import build_sequences
from osa.paths import analysis_path
from osa.report import start
from osa.utils.cliopts import sequencer_cli_parsing
from osa.utils.logging import myLogger
from osa.utils.utils import is_day_closed, gettag, date_to_iso
from osa.veto import get_closed_list, get_veto_list

__all__ = [
    "single_process",
    "update_sequence_status",
    "get_status_for_sequence",
    "output_matrix",
    "report_sequences",
    "update_job_info",
]

log = myLogger(logging.getLogger())


def main():
    """
    Main script to be called as cron job. It creates and execute
    the calibration sequence and afterward it prepares a SLURM job
    array which launches the data sequences for every subrun.
    """
    sequencer_cli_parsing()

    if options.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    single_array = ["LST1", "LST2"]
    tag = gettag()
    start(tag)
    if options.tel_id in single_array:
        single_process(options.tel_id)
    else:
        log.error("Process mode not supported yet")


def single_process(telescope):
    """
    Runs the single process for a single telescope

    Parameters
    ----------
    telescope : str
        Options: 'LST1'

    Returns
    -------
    sequence_list : list
    """

    database = cfg.get("database", "path")
    if database:
        osadb.start_processing(date_to_iso(options.date))

    # Define global variables and create night directory
    sequence_list = []
    options.tel_id = telescope
    options.directory = analysis_path(options.tel_id)
    options.log_directory = options.directory / "log"

    if not options.simulate:
        os.makedirs(options.log_directory, exist_ok=True)

    if is_day_closed():
        log.info(f"Date {date_to_iso(options.date)} is already closed for {options.tel_id}")
        return sequence_list

    # Build the sequences
    sequence_list = build_sequences(options.date)

    # Create job pilot scripts
    prepare_jobs(sequence_list)

    # Update sequences objects with information from SLURM
    update_job_info(sequence_list)

    get_veto_list(sequence_list)
    get_closed_list(sequence_list)
    update_sequence_status(sequence_list)

    if not options.no_submit:
        submit_jobs(sequence_list)

    # TODO: insert_new_activity_db(sequence_list)

    # Display the sequencer table with processing status
    report_sequences(sequence_list)

    return sequence_list


def update_job_info(sequence_list):
    """
    Updates the job information from SLURM

    Parameters
    ----------
    sequence_list : list
        List of sequences to be updated
    """
    if options.test:
        return

    sacct_output, squeue_output = run_sacct(), run_squeue()
    set_queue_values(
        sacct_info=get_sacct_output(sacct_output),
        squeue_info=get_squeue_output(squeue_output),
        sequence_list=sequence_list,
    )


def update_sequence_status(seq_list):
    """
    Update the percentage of files produced of each type (calibration, DL1,
    DATACHECK, MUON and DL2) for every run considering the total number of subruns.

    Parameters
    ----------
    seq_list
        List of sequences of a given night corresponding to each run.
    """
    for seq in seq_list:
        if seq.type == "PEDCALIB":
            seq.calibstatus = int(
                Decimal(get_status_for_sequence(seq, "CALIB") * 100) / seq.subruns
            )
        elif seq.type == "DATA":
            seq.dl1status = int(Decimal(get_status_for_sequence(seq, "DL1") * 100) / seq.subruns)
            seq.dl1abstatus = int(
                Decimal(get_status_for_sequence(seq, "DL1AB") * 100) / seq.subruns
            )
            seq.datacheckstatus = int(
                Decimal(get_status_for_sequence(seq, "DATACHECK") * 100) / seq.subruns
            )
            seq.muonstatus = int(Decimal(get_status_for_sequence(seq, "MUON") * 100) / seq.subruns)
            seq.dl2status = int(Decimal(get_status_for_sequence(seq, "DL2") * 100) / seq.subruns)


def get_status_for_sequence(sequence, data_level) -> int:
    """
    Get number of files produced for a given sequence and data level.

    Parameters
    ----------
    sequence
    data_level : str
        Options: 'CALIB', 'DL1', 'DL1AB', 'DATACHECK', 'MUON' or 'DL2'

    Returns
    -------
    number_of_files : int
    """
    if data_level == "DL1AB":
        directory = options.directory / options.dl1_prod_id
        files = list(directory.glob(f"dl1_LST-1*{sequence.run}*.h5"))

    elif data_level == "DL2":
        directory = options.directory / options.dl2_prod_id
        files = list(directory.glob(f"dl2_LST-1*{sequence.run}*.h5"))

    elif data_level == "DATACHECK":
        directory = options.directory / options.dl1_prod_id
        files = list(directory.glob(f"datacheck_dl1_LST-1*{sequence.run}*.h5"))

    else:
        prefix = cfg.get("PATTERN", f"{data_level}PREFIX")
        suffix = cfg.get("PATTERN", f"{data_level}SUFFIX")
        files = list(options.directory.glob(f"{prefix}*{sequence.run}*{suffix}"))

    return len(files)


def report_sequences(sequence_list):
    """
    Update the status report table shown by the sequencer.

    Parameters
    ----------
    sequence_list: list
        List of sequences of a given date
    """
    header = [
        "Tel",
        "Seq",
        "Parent",
        "Type",
        "Run",
        "Subruns",
        "Source",
        "Action",
        "Tries",
        "JobID",
        "State",
        "CPU_time",
        "Exit",
    ]
    if options.tel_id in ["LST1", "LST2"]:
        header.extend(("DL1%", "MUONS%", "DL1AB%", "DATACHECK%", "DL2%"))
    matrix = [header]
    for sequence in sequence_list:
        row_list = [
            sequence.telescope,
            sequence.seq,
            sequence.parent,
            sequence.type,
            sequence.run,
            sequence.subruns,
            sequence.source_name,
            sequence.action,
            sequence.tries,
            sequence.jobid,
            sequence.state,
            sequence.cputime,
            sequence.exit,
        ]
        if sequence.type in ["DRS4", "PEDCALIB"]:
            row_list.extend((None, None, None, None, None))
        elif sequence.type == "DATA":
            row_list.extend(
                (
                    sequence.dl1status,
                    sequence.muonstatus,
                    sequence.dl1abstatus,
                    sequence.datacheckstatus,
                    sequence.dl2status,
                )
            )

        matrix.append(row_list)
    padding = int(cfg.get("OUTPUT", "PADDING"))
    output_matrix(matrix, padding)


def output_matrix(matrix: list, padding_space: int):
    """
    Build the status table shown by the sequencer.

    Parameters
    ----------
    matrix: list
    padding_space: int
    """
    max_field_length = []
    for row in matrix:
        for j, col in enumerate(row):
            if matrix.index(row) == 0:
                max_field_length.append(len(str(col)))
            elif len(str(col)) > max_field_length[j]:
                # Insert or update the first length
                max_field_length[j] = len(str(col))
    for row in matrix:
        stringrow = ""
        rpadding = padding_space * " "
        for j, col in enumerate(row):
            lpadding = (max_field_length[j] - len(str(col))) * " "
            stringrow += (
                f"{lpadding}{col}{rpadding}"
                if isinstance(col, int)
                else f"{col}{lpadding}{rpadding}"
            )

        log.info(stringrow)


if __name__ == "__main__":
    main()



"""Script to run the gain selection over a list of dates."""
import logging
import re
import shutil
import glob
import pandas as pd
import subprocess as sp
from pathlib import Path
from textwrap import dedent
from io import StringIO

import click
from astropy.table import Table
from lstchain.paths import run_info_from_filename, parse_r0_filename

from osa.scripts.reprocessing import get_list_of_dates, check_job_status_and_wait
from osa.utils.utils import wait_for_daytime
from osa.utils.logging import myLogger
from osa.job import get_sacct_output, FORMAT_SLURM

log = myLogger(logging.getLogger(__name__))

PATH = "PATH=/fefs/aswg/software/offline_dvr/bin:$PATH"


def get_sbatch_script(
    run_id, subrun, input_file, output_dir, log_dir, ref_time, ref_counter, module, ref_source
):
    """Build the sbatch job pilot script for running the gain selection."""
    return dedent(
        f"""\
    #!/bin/bash

    #SBATCH -D {log_dir}
    #SBATCH -o "gain_selection_{run_id:05d}_{subrun:04d}_%j.log"
    #SBATCH --job-name "gain_selection_{run_id:05d}"
    #SBATCH --export {PATH}

    lst_dvr {input_file} {output_dir} {ref_time} {ref_counter} {module} {ref_source}
    """
    )


def apply_gain_selection(date: str, start: int, end: int, output_basedir: Path = None):
    """
    Submit the jobs to apply the gain selection to the data for a given date
    on a subrun-by-subrun basis.
    """

    run_summary_dir = Path("/fefs/aswg/data/real/monitoring/RunSummary")
    run_summary_file = run_summary_dir / f"RunSummary_{date}.ecsv"
    summary_table = Table.read(run_summary_file)
    # Apply gain selection only to DATA runs
    data_runs = summary_table[summary_table["run_type"] == "DATA"]

    output_dir = output_basedir / date
    log_dir = output_basedir / "log" / date
    output_dir.mkdir(parents=True, exist_ok=True)
    log_dir.mkdir(parents=True, exist_ok=True)
    r0_dir = Path(f"/fefs/aswg/data/real/R0/{date}")

    for run in data_runs:
        # Check slurm queue status and sleep for a while to avoid overwhelming the queue
        check_job_status_and_wait(max_jobs=1500)

        # Avoid running jobs while it is still night time
        wait_for_daytime(start, end)

        run_id = run["run_id"]
        ref_time = run["dragon_reference_time"]
        ref_counter = run["dragon_reference_counter"]
        module = run["dragon_reference_module_index"]
        ref_source = run["dragon_reference_source"].upper()

        files = glob.glob(f"{r0_dir}/LST-1.?.Run{run_id:05d}.????.fits.fz")
        subrun_numbers = [int(file[-12:-8]) for file in files]
        input_files = []

        if ref_source in ["UCTS", "TIB"]:

            n_subruns = max(subrun_numbers)

            for subrun in range(n_subruns + 1):
                new_files = glob.glob(f"{r0_dir}/LST-1.?.Run{run_id:05d}.{subrun:04d}.fits.fz")

                if len(new_files) != 4:
                    for file in new_files:
                        sp.run(["cp", file, output_dir])

                else:
                    new_files.sort()
                    input_files.append(new_files[0])

            for file in input_files:
                run_info = run_info_from_filename(file)
                job_file = log_dir / f"gain_selection_{run_info.run:05d}.{run_info.subrun:04d}.sh"
                with open(job_file, "w") as f:
                    f.write(
                        get_sbatch_script(
                            run_id,
                            run_info.subrun,
                            file,
                            output_dir,
                            log_dir,
                            ref_time,
                            ref_counter,
                            module,
                            ref_source,
                        )
                    )
                sp.run(["sbatch", job_file], check=True)

        else:

            input_files = r0_dir.glob(f"LST-1.?.Run{run_id:05d}.????.fits.fz")

            for file in input_files:
                sp.run(["cp", file, output_dir])

    calib_runs = summary_table[summary_table["run_type"] != "DATA"]

    for run in calib_runs:
        # Avoid copying files while it is still night time
        wait_for_daytime(start, end)

        run_id = run["run_id"]
        r0_files = r0_dir.glob(f"LST-1.?.Run{run_id:05d}.????.fits.fz")

        for file in r0_files:
            sp.run(["cp", file, output_dir])

def run_sacct_j(job) -> StringIO:
    """Run sacct to obtain the job information."""
    if shutil.which("sacct") is None:
        log.warning("No job info available since sacct command is not available")
        return StringIO()

    sacct_cmd = [
        "sacct",
        "-n",
        "--parsable2",
        "--delimiter=,",
        "--units=G",
        "-o",
        ",".join(FORMAT_SLURM),
        "-j",
        job,
    ]

    return StringIO(sp.check_output(sacct_cmd).decode())


def check_failed_jobs(date: str, output_basedir: Path = None):
    """Search for failed jobs in the log directory."""
    failed_jobs = []
    log_dir = output_basedir / "log" / date
    filenames = glob.glob(f"{log_dir}/gain_selection*.log")
    jobs = [re.search(r'(?<=_)(.[0-9.]+?)(?=.log)', i).group(0) for i in filenames]

    for job in jobs:
        output = run_sacct_j(job)
        df = get_sacct_output(output)

        if not df.iloc[0]["State"] == "COMPLETED":
            log.warning(f"Job {job} did not finish successfully")
            failed_jobs.append(job)

    if not failed_jobs:
        log.info(f"{date}: all jobs finished successfully")
    else:
        log.warning(f"{date}: some jobs did not finish successfully")


    run_summary_dir = Path("/fefs/aswg/data/real/monitoring/RunSummary")
    run_summary_file = run_summary_dir / f"RunSummary_{date}.ecsv"
    summary_table = Table.read(run_summary_file)
    runs = summary_table["run_id"]
    missing_runs = []

    r0_files = glob.glob(f"/fefs/aswg/data/real/R0/{date}/LST-1.?.Run?????.????.fits.fz")
    r0g_files = glob.glob(f"/fefs/aswg/data/real/R0G/{date}/LST-1.?.Run?????.????.fits.fz")
    all_r0_runs = [parse_r0_filename(i).run for i in r0_files]
    all_r0g_runs = [parse_r0_filename(i).run for i in r0g_files]

    for run in all_r0_runs:
        if run not in runs:
            if run not in all_r0g_runs:
                missing_runs.append(run)

    missing_runs.sort()
    if missing_runs:
        log.info(
            f"Some runs are missing. Copying R0 files of runs {pd.Series(missing_runs).unique()} "
            f"directly to /fefs/aswg/data/real/R0G/{date}"
        )

        for run in missing_runs:
            output_dir = Path(f"/fefs/aswg/data/real/R0G/{date}/")
            files = glob.glob(f"/fefs/aswg/data/real/R0/{date}/LST-1.?.Run{run:05d}.????.fits.fz")
            for file in files:
                sp.run(["cp", file, output_dir])



@click.command()
@click.option("--check", is_flag=True, default=False, help="Check for failed jobs.")
@click.argument("dates-file", type=click.Path(exists=True, path_type=Path))
@click.argument("output-basedir", type=click.Path(path_type=Path))
@click.option("-s", "--start-time", type=int, default=10, help="Time to (re)start gain selection in HH format.")
@click.option("-e", "--end-time", type=int, default=18, help="Time to stop gain selection in HH format.")
def main(
    dates_file: Path = None, 
    output_basedir: Path = None, 
    check: bool = False, 
    start_time: int = 10, 
    end_time: int = 18
):
    """
    Loop over the dates listed in the input file and launch the gain selection
    script for each of them. The input file should list the dates in the format
    YYYYMMDD one date per line.
    """
    log.setLevel(logging.INFO)

    list_of_dates = get_list_of_dates(dates_file)

    if check:
        for date in list_of_dates:
            check_failed_jobs(date, output_basedir)
    else:
        for date in list_of_dates:
            apply_gain_selection(date, start_time, end_time, output_basedir)
        log.info("Done! No more dates to process.")


if __name__ == "__main__":
    main()



"""Functions to handle the interaction with the job scheduler."""

import datetime
import logging
import shutil
import subprocess as sp
import time
from io import StringIO
from pathlib import Path
from textwrap import dedent
from typing import Iterable

import matplotlib.pyplot as plt
import pandas as pd

from osa.configs import options
from osa.configs.config import cfg
from osa.paths import (
    pedestal_ids_file_exists,
    get_drive_file,
    get_summary_file,
    get_pedestal_ids_file,
)
from osa.utils.iofile import write_to_file
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_dir, time_to_seconds, stringify, date_to_iso

log = myLogger(logging.getLogger(__name__))

__all__ = [
    "are_all_jobs_correctly_finished",
    "historylevel",
    "prepare_jobs",
    "sequence_filenames",
    "set_queue_values",
    "job_header_template",
    "plot_job_statistics",
    "scheduler_env_variables",
    "set_cache_dirs",
    "submit_jobs",
    "check_history_level",
    "get_sacct_output",
    "get_squeue_output",
    "filter_jobs",
    "run_sacct",
    "run_squeue",
    "calibration_sequence_job_template",
    "data_sequence_job_template",
    "save_job_information",
]

TAB = "\t".expandtabs(4)
FORMAT_SLURM = [
    "JobID",
    "JobName",
    "CPUTime",
    "CPUTimeRAW",
    "Elapsed",
    "TotalCPU",
    "MaxRSS",
    "State",
    "ExitCode",
]

PYTHON_IMPORTS = dedent(
    """\

    import os
    import subprocess
    import sys
    import tempfile

    """
)


def are_all_jobs_correctly_finished(sequence_list):
    """
    Check if all jobs are correctly finished by looking
    at the history file.

    Parameters
    ----------
    sequence_list: list
        List of sequence objects

    Returns
    -------
    flag: bool
    """
    # FIXME: check based on sequence.jobid exit status
    flag = True
    analysis_directory = Path(options.directory)
    for sequence in sequence_list:
        history_files_list = analysis_directory.rglob(f"*{sequence.seq}*.history")
        for history_file in history_files_list:
            # TODO: s.history should be SubRunObj attribute not RunObj
            # s.history only working for CALIBRATION sequence (run-wise), since it is
            # looking for .../sequence_LST1_04180.history files
            # we need to check all the subrun wise history files
            # .../sequence_LST1_04180.XXXX.history
            out, _ = historylevel(history_file, sequence.type)
            if out == 0:
                log.debug(f"Job {sequence.seq} ({sequence.type}) correctly finished")
                continue
            if out == 1 and options.no_dl2:
                log.debug(
                    f"Job {sequence.seq} ({sequence.type}) correctly "
                    f"finished up to DL1ab, but --no-dl2 option selected"
                )
                continue

            log.warning(
                f"Job {sequence.seq} (run {sequence.run}) not correctly finished [level {out}]"
            )
            flag = False
    return flag


def check_history_level(history_file: Path, program_levels: dict):
    """
    Check the history of the calibration sequence.

    Parameters
    ----------
    history_file: pathlib.Path
        Path to the history file
    program_levels: dict
        Dictionary with the program name and the level of the program

    Returns
    -------
    level: int
        Level of the history file
    exit_status: int
        Exit status pf the program according to the history file
    """

    # Check the program exit_status (last string of the line), if it is 0, go
    # to the next level and check the next program. If exit_status is not 0, return the
    # actual level and the exit status. Stop the iteration when reaching the level 0.
    with open(history_file, "r") as file:
        for line in file:
            program = line.split()[1]
            exit_status = int(line.split()[-1])
            if program in program_levels:
                if exit_status != 0:
                    level = program_levels[program]
                    return level, exit_status
                level = program_levels[program]
                continue

        return level, exit_status


def historylevel(history_file: Path, data_type: str):
    """
    Returns the level from which the analysis should begin and
    the rc of the last executable given a certain history file.

    Notes
    -----
    Workflow for PEDCALIB sequences:
     - Creation of DRS4 pedestal file, level 2->1
     - Creation of charge calibration file, level 1->0
     - Sequence completed when reaching level 0

    Workflow for DATA sequences:
     - R0->DL1, level 4->3
     - DL1->DL1AB, level 3->2
     - DATACHECK, level 2->1
     - DL1->DL2, level 1->0
     - Sequence completed when reaching level 0

    Parameters
    ----------
    history_file: pathlib.Path
    data_type: str
        Type of the sequence, either 'DATA' or 'PEDCALIB'

    Returns
    -------
    level : int
    exit_status : int
    """

    # TODO: Create a dict with the program exit status and prod id to take
    #  into account not only the last history line but also the others.

    if data_type == "DATA":
        level = 4
    elif data_type == "PEDCALIB":
        level = 2
    else:
        raise ValueError(f"Type {data_type} not expected")

    exit_status = 0

    if history_file.exists():
        for line in history_file.read_text().splitlines():
            words = line.split()
            try:
                program = words[1]
                prod_id = words[2]
                exit_status = int(words[-1])
                log.debug(f"{program}, finished with error {exit_status} and prod ID {prod_id}")
            except (IndexError, ValueError) as err:
                log.exception(f"Malformed history file {history_file}, {err}")
            else:
                # Calibration sequence
                if program == cfg.get("lstchain", "drs4_baseline"):
                    level = 1 if exit_status == 0 else 2
                elif program == cfg.get("lstchain", "charge_calibration"):
                    level = 0 if exit_status == 0 else 1
                # Data sequence
                elif program == cfg.get("lstchain", "r0_to_dl1"):
                    level = 3 if exit_status == 0 else 4
                elif program == cfg.get("lstchain", "dl1ab"):
                    if (exit_status == 0) and (prod_id == options.dl1_prod_id):
                        log.debug(f"DL1ab prod ID: {options.dl1_prod_id} already produced")
                        level = 2
                    else:
                        level = 3
                        log.debug(f"DL1ab prod ID: {options.dl1_prod_id} not produced yet")
                        break
                elif program == cfg.get("lstchain", "check_dl1"):
                    level = 1 if exit_status == 0 else 2
                elif program == cfg.get("lstchain", "dl1_to_dl2"):
                    if (exit_status == 0) and (prod_id == options.dl2_prod_id):
                        log.debug(f"DL2 prod ID: {options.dl2_prod_id} already produced")
                        level = 0
                    else:
                        level = 1
                        log.debug(f"DL2 prod ID: {options.dl2_prod_id} not produced yet")

                else:
                    log.warning(f"Program name not identified: {program}")

    return level, exit_status


def prepare_jobs(sequence_list):
    """Prepare job file template for each sequence."""
    if not options.simulate:
        log.info("Building job scripts for each sequence.")

    for sequence in sequence_list:
        log.debug(f"Creating sequence.py for sequence {sequence.seq}")
        if sequence.type == "PEDCALIB":
            calibration_sequence_job_template(sequence)
        elif sequence.type == "DATA":
            data_sequence_job_template(sequence)
        else:
            raise ValueError(f"Type {sequence.type} not expected")


def sequence_filenames(sequence):
    """Build names of the script, veto and history files."""
    basename = f"sequence_{sequence.jobname}"
    sequence.script = Path(options.directory) / f"{basename}.py"
    sequence.veto = Path(options.directory) / f"{basename}.veto"
    sequence.history = Path(options.directory) / f"{basename}.history"


def save_job_information():
    """
    Write job information from sacct (elapsed time, memory used, number of
    completed, failed and running jobs in the queue) to a file.
    """
    # Set directory and file path
    log_directory = Path(options.directory) / "log"
    log_directory.mkdir(exist_ok=True, parents=True)
    file_path = log_directory / "job_information.csv"

    sacct_output = run_sacct()
    jobs_df = get_sacct_output(sacct_output)

    # Fetch sacct output and prepare the data
    jobs_df_filtered = jobs_df.copy()
    jobs_df_filtered = jobs_df_filtered.dropna()
    # Remove the G from MaxRSS value and convert to float
    # jobs_df_filtered["MaxRSS"] = jobs_df_filtered["MaxRSS"].str.strip("G").astype(float)

    jobs_df_filtered.to_csv(file_path, index=False, sep=",")


def plot_job_statistics(sacct_output: pd.DataFrame, directory: Path):
    """
    Get statistics of the jobs. Check elapsed time used,
    the memory used, the number of jobs completed, the number of jobs failed,
    the number of jobs running, the number of jobs queued.
    It will fetch the information from the sacct output.

    Parameters
    ----------
    sacct_output: pd.DataFrame
    directory: Path
        Directory to save the plot.
    """
    # TODO: this function will be called in the closer loop after all
    #  the jobs are done for a given production.

    # Plot a 2D histogram of the used memory (MaxRSS) as a function of the
    # elapsed time taking also into account the State of the job.
    sacct_output_filter = sacct_output.copy()
    sacct_output_filter = sacct_output_filter.dropna()
    # Remove the G from MaxRSS value and convert to float
    sacct_output_filter["MaxRSS"] = sacct_output_filter["MaxRSS"].str.strip("G").astype(float)

    plt.figure()
    plt.hist2d(sacct_output_filter.MaxRSS, sacct_output_filter.CPUTimeRAW / 3600, bins=50)
    plt.xlabel("MaxRSS [GB]")
    plt.ylabel("Elapsed time [h]")
    directory.mkdir(exist_ok=True, parents=True)
    plot_path = directory / "job_statistics.pdf"
    plt.savefig(plot_path)


def scheduler_env_variables(sequence, scheduler="slurm"):
    """Return the environment variables for the scheduler."""
    # TODO: Create a class with the SBATCH variables we want to use in the pilot job
    #  and then use the string representation of the class to create the header.
    if scheduler != "slurm":
        log.warning("No other schedulers are currently supported")
        return None

    sbatch_parameters = [
        f"--job-name={sequence.jobname}",
        f"--time={cfg.get('SLURM', 'WALLTIME')}",
        f"--chdir={options.directory}",
        f"--output=log/Run{sequence.run:05d}.%4a_jobid_%A.out",
        f"--error=log/Run{sequence.run:05d}.%4a_jobid_%A.err",
    ]

    # Get the number of subruns counting from 0.
    subruns = sequence.subruns - 1

    # Depending on the type of sequence, we need to set
    # different sbatch environment variables
    if sequence.type == "DATA":
        sbatch_parameters.append(f"--array=0-{subruns}")

    sbatch_parameters.append(f"--partition={cfg.get('SLURM', f'PARTITION_{sequence.type}')}")
    sbatch_parameters.append(f"--mem-per-cpu={cfg.get('SLURM', f'MEMSIZE_{sequence.type}')}")

    return ["#SBATCH " + line for line in sbatch_parameters]


def job_header_template(sequence):
    """
    Returns a string with the job header template
    including SBATCH environment variables for sequencerXX.py script

    Parameters
    ----------
    sequence: sequence object

    Returns
    -------
    header: str
        String with job header template
    """
    python_shebang = "#!/bin/env python"
    if options.test:
        return python_shebang
    sbatch_parameters = "\n".join(scheduler_env_variables(sequence))
    return python_shebang + 2 * "\n" + sbatch_parameters


def set_cache_dirs():
    """
    Export cache directories for the jobs provided they
    are defined in the config file.

    Returns
    -------
    content: string
        String with the command to export the cache directories
    """

    ctapipe_cache = cfg.get("CACHE", "CTAPIPE_CACHE")
    ctapipe_svc_path = cfg.get("CACHE", "CTAPIPE_SVC_PATH")
    mpl_config_path = cfg.get("CACHE", "MPLCONFIGDIR")

    content = []
    if ctapipe_cache:
        content.append(f"os.environ['CTAPIPE_CACHE'] = '{ctapipe_cache}'")

    if ctapipe_svc_path:
        content.append(f"os.environ['CTAPIPE_SVC_PATH'] = '{ctapipe_svc_path}'")

    if mpl_config_path:
        content.append(f"os.environ['MPLCONFIGDIR'] = '{mpl_config_path}'")

    return "\n".join(content)


def data_sequence_job_template(sequence):
    """
    This file contains instruction to be submitted to job scheduler.

    Parameters
    ----------
    sequence : sequence object

    Returns
    -------
    job_template : string
    """
    # TODO: refactor this function creating wrappers that handle slurm part

    # Get the job header template.
    job_header = job_header_template(sequence)

    flat_date = date_to_dir(options.date)

    commandargs = ["datasequence"]

    if options.verbose:
        commandargs.append("-v")
    if options.simulate:
        commandargs.append("-s")
    if options.configfile:
        commandargs.extend(("--config", f"{Path(options.configfile).resolve()}"))
    if sequence.type == "DATA" and options.no_dl2:
        commandargs.append("--no-dl2")

    commandargs.extend(
        (
            f"--date={date_to_iso(options.date)}",
            f"--prod-id={options.prod_id}",
            f"--drs4-pedestal-file={sequence.drs4_file}",
            f"--time-calib-file={sequence.time_calibration_file}",
            f"--pedcal-file={sequence.calibration_file}",
            f"--systematic-correction-file={sequence.systematic_correction_file}",
            f"--drive-file={get_drive_file(flat_date)}",
            f"--run-summary={get_summary_file(flat_date)}",
        )
    )

    content = job_header + "\n" + PYTHON_IMPORTS

    if not options.test:
        content += set_cache_dirs()
        content += "\n"
        # Use the SLURM env variables
        content += "subruns = int(os.getenv('SLURM_ARRAY_TASK_ID'))\n"
    else:
        # Just process the first subrun without SLURM
        content += "subruns = 0\n"

    content += "\n"

    content += "with tempfile.TemporaryDirectory() as tmpdirname:\n"
    content += TAB + "os.environ['NUMBA_CACHE_DIR'] = tmpdirname\n"

    content += TAB + "proc = subprocess.run([\n"

    for arg in commandargs:
        content += TAB * 2 + f"'{arg}',\n"

    if pedestal_ids_file_exists(sequence.run):
        pedestal_ids_file = get_pedestal_ids_file(sequence.run, flat_date)
        content += TAB * 2 + f"f'--pedestal-ids-file={pedestal_ids_file}',\n"

    content += TAB * 2 + f"f'{sequence.run:05d}.{{subruns:04d}}',\n"

    content += TAB * 2 + f"'{options.tel_id}'\n"
    content += TAB + "])\n"
    content += "\n"
    content += "sys.exit(proc.returncode)"

    if not options.simulate:
        write_to_file(sequence.script, content)

    return content


def calibration_sequence_job_template(sequence):
    """
    This file contains instruction to be submitted to job scheduler.

    Parameters
    ----------
    sequence : sequence object

    Returns
    -------
    job_template : string
    """

    # Get the job header template.
    job_header = job_header_template(sequence)

    commandargs = ["calibration_pipeline"]

    if options.verbose:
        commandargs.append("-v")
    if options.simulate:
        commandargs.append("-s")
    if options.configfile:
        commandargs.extend(("--config", f"{Path(options.configfile).resolve()}"))
    commandargs.extend(
        (
            f"--date={date_to_iso(options.date)}",
            f"--drs4-pedestal-run={sequence.drs4_run:05d}",
            f"--pedcal-run={sequence.run:05d}",
        )
    )

    content = job_header + "\n" + PYTHON_IMPORTS

    if not options.test:
        content += set_cache_dirs()
        content += "\n"
        # Use the SLURM env variables
        content += "subruns = os.getenv('SLURM_ARRAY_TASK_ID')\n"
    else:
        # Just process the first subrun without SLURM
        content += "subruns = 0\n"

    content += "\n"

    content += "with tempfile.TemporaryDirectory() as tmpdirname:\n"
    content += TAB + "os.environ['NUMBA_CACHE_DIR'] = tmpdirname\n"

    content += TAB + "proc = subprocess.run([\n"

    for arg in commandargs:
        content += TAB * 2 + f"'{arg}',\n"

    content += TAB * 2 + f"'{options.tel_id}'\n"
    content += TAB + "])\n"
    content += "\n"
    content += "sys.exit(proc.returncode)"

    if not options.simulate:
        write_to_file(sequence.script, content)

    return content


def submit_jobs(sequence_list, batch_command="sbatch"):
    """
    Submit the jobs to the cluster.

    Parameters
    ----------
    sequence_list: list
        List of sequences to submit.
    batch_command: str
        The batch command to submit the job (Default: sbatch)

    Returns
    -------
    job_list: list
        List of submitted job IDs.
    """
    job_list = []
    no_display_backend = "--export=ALL,MPLBACKEND=Agg"

    for sequence in sequence_list:
        commandargs = [batch_command, "--parsable", no_display_backend]
        if sequence.type == "PEDCALIB":
            commandargs.append(str(sequence.script))
            if options.simulate or options.no_calib or options.test:
                log.debug("SIMULATE Launching scripts")
            else:
                try:
                    log.debug(f"Launching script {sequence.script}")
                    parent_jobid = sp.check_output(
                        commandargs, universal_newlines=True, shell=False
                    ).split()[0]
                except sp.CalledProcessError as error:
                    rc = error.returncode
                    log.exception(f"Command '{batch_command}' not found, error {rc}")

            log.debug(stringify(commandargs))

        # Here sequence.jobid has not been redefined, so it keeps the one
        # from previous time sequencer was launched.

        # Add the job dependencies after calibration sequence
        if sequence.type == "DATA":
            if not options.simulate and not options.no_calib and not options.test:
                log.debug("Adding dependencies to job submission")
                depend_string = f"--dependency=afterok:{parent_jobid}"
                commandargs.append(depend_string)

            commandargs.append(sequence.script)

            if options.simulate:
                log.debug("SIMULATE Launching scripts")
            elif options.test:
                log.debug(
                    "TEST launching datasequence scripts for " "first subrun without scheduler"
                )
                commandargs = ["python", sequence.script]
                sp.check_output(commandargs, shell=False)
            else:
                log.info("Submitting jobs to the cluster.")
                try:
                    log.debug(f"Launching script {sequence.script}")
                    sp.check_output(commandargs, shell=False)
                except sp.CalledProcessError as error:
                    log.exception(error)

            log.debug(stringify(commandargs))

        job_list.append(sequence.script)

    return job_list


def run_squeue() -> StringIO:
    """Run squeue command to get the status of the jobs."""
    if shutil.which("squeue") is None:
        log.warning("No job info available since squeue command is not available")
        return StringIO()

    out_fmt = "%i;%j;%T;%M"  # JOBID, NAME, STATE, TIME
    return StringIO(sp.check_output(["squeue", "--me", "-o", out_fmt]).decode())


def get_squeue_output(squeue_output: StringIO) -> pd.DataFrame:
    """
    Obtain the current job information from squeue output
    and return a pandas dataframe.
    """
    df = pd.read_csv(squeue_output, delimiter=";")
    df.rename(
        inplace=True,
        columns={
            "STATE": "State",
            "JOBID": "JobID",
            "NAME": "JobName",
            "TIME": "CPUTime",
        },
    )

    # Keep only the jobs corresponding to OSA sequences
    df = df[df["JobName"].str.contains("LST1")]

    try:
        # Remove the job array part of the jobid
        df["JobID"] = df["JobID"].apply(lambda x: x.split("_")[0]).astype("int")
    except AttributeError:
        log.debug("No job info could be obtained from squeue")

    df["CPUTimeRAW"] = df["CPUTime"].apply(time_to_seconds)

    return df


def run_sacct() -> StringIO:
    """Run sacct to obtain the job information."""
    if shutil.which("sacct") is None:
        log.warning("No job info available since sacct command is not available")
        return StringIO()

    sacct_cmd = [
        "sacct",
        "-n",
        "--parsable2",
        "--delimiter=,",
        "--units=G",
        "-o",
        ",".join(FORMAT_SLURM),
    ]
    if cfg.get("SLURM", "STARTTIME_DAYS_SACCT"):
        days = int(cfg.get("SLURM", "STARTTIME_DAYS_SACCT"))
        start_date = (datetime.date.today() - datetime.timedelta(days=days)).isoformat()
        sacct_cmd.extend(["--starttime", start_date])

    return StringIO(sp.check_output(sacct_cmd).decode())


def get_sacct_output(sacct_output: StringIO) -> pd.DataFrame:
    """
    Fetch the information of jobs in the queue using the sacct SLURM output
    and store it in a pandas dataframe.

    Returns
    -------
    queue_list: pd.DataFrame
    """
    sacct_output = pd.read_csv(sacct_output, names=FORMAT_SLURM)

    # Keep only the jobs corresponding to OSA sequences
    sacct_output = sacct_output[
        (sacct_output["JobName"].str.contains("batch"))
        | (sacct_output["JobName"].str.contains("LST1"))
    ]

    try:
        sacct_output["JobID"] = sacct_output["JobID"].apply(lambda x: x.split("_")[0])
        sacct_output["JobID"] = sacct_output["JobID"].str.strip(".batch").astype(int)

    except AttributeError:
        log.debug("No job info could be obtained from sacct")

    return sacct_output


def filter_jobs(job_info: pd.DataFrame, sequence_list: Iterable):
    """Filter the job info list to get the values of the jobs in the current queue."""
    sequences_info = pd.DataFrame([vars(seq) for seq in sequence_list])
    # Keep the jobs in the sacct output that are present in the sequence list
    return job_info[job_info["JobName"].isin(sequences_info["jobname"])]


def set_queue_values(
    sacct_info: pd.DataFrame, squeue_info: pd.DataFrame, sequence_list: Iterable
) -> None:
    """
    Extract job info from sacct output and
    fetch them into the table of sequences.

    Parameters
    ----------
    sacct_info: pd.DataFrame
    squeue_info: pd.DataFrame
    sequence_list: list[Sequence object]
    """
    if sacct_info.empty and squeue_info.empty or sequence_list is None:
        return

    job_info = pd.concat([sacct_info, squeue_info])

    # Filter the jobs in the sacct output that are present in the sequence list
    job_info_filtered = filter_jobs(job_info, sequence_list)

    for sequence in sequence_list:
        df_jobname = job_info_filtered[job_info_filtered["JobName"] == sequence.jobname]
        sequence.tries = df_jobname["JobID"].nunique()
        sequence.action = "Check"

        if not df_jobname.empty:
            sequence.jobid = df_jobname["JobID"].max()  # Get latest JobID
            df_jobid_filtered = df_jobname[df_jobname["JobID"] == sequence.jobid]
            try:
                sequence.cputime = time.strftime(
                    "%H:%M:%S",
                    time.gmtime(df_jobid_filtered["CPUTimeRAW"].median(skipna=False)),
                )
            except ValueError:
                sequence.cputime = None

            update_sequence_state(sequence, df_jobid_filtered)


def update_sequence_state(sequence, filtered_job_info: pd.DataFrame) -> None:
    """
    Update the state of the sequence based on the job info.

    Parameters
    ----------
    sequence: Sequence object
    filtered_job_info: pd.DataFrame
    """
    if (filtered_job_info.State.values == "COMPLETED").all():
        sequence.state = "COMPLETED"
        sequence.exit = filtered_job_info["ExitCode"].iloc[0]
    elif (filtered_job_info.State.values == "PENDING").all():
        sequence.state = "PENDING"
    elif any("FAILED" in job for job in filtered_job_info.State):
        sequence.state = "FAILED"
        sequence.exit = filtered_job_info[filtered_job_info.State.values == "FAILED"][
            "ExitCode"
        ].iloc[0]
    elif any("CANCELLED" in job for job in filtered_job_info.State):
        sequence.state = "CANCELLED"
        mask = ["CANCELLED" in job for job in filtered_job_info.State]
        sequence.exit = filtered_job_info[mask]["ExitCode"].iloc[0]
    elif any("TIMEOUT" in job for job in filtered_job_info.State):
        sequence.state = "TIMEOUT"
        sequence.exit = "0:15"
    elif any("RUNNING" in job for job in filtered_job_info.State):
        sequence.state = "RUNNING"



"""Command line argument parser for lstosa scripts."""

import datetime
import logging
from argparse import ArgumentParser
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.paths import analysis_path, DEFAULT_CFG
from osa.utils.logging import myLogger
from osa.utils.utils import (
    get_dl1_prod_id,
    get_dl2_prod_id,
    get_prod_id,
    is_defined,
    set_prod_ids,
    YESTERDAY,
)

__all__ = [
    "closer_argparser",
    "closercliparsing",
    "copy_datacheck_parsing",
    "data_sequence_cli_parsing",
    "data_sequence_argparser",
    "provprocessparsing",
    "sequencer_argparser",
    "sequencer_cli_parsing",
    "set_default_date_if_needed",
    "simprocparsing",
    "sequencer_webmaker_argparser",
    "valid_date",
    "get_prod_id",
    "get_dl1_prod_id",
    "get_dl2_prod_id",
    "calibration_pipeline_cliparsing",
    "calibration_pipeline_argparser",
    "autocloser_cli_parser",
    "common_parser",
]

log = myLogger(logging.getLogger(__name__))


def valid_date(string):
    """Check if the string is a valid date and return a datetime object."""
    return datetime.datetime.strptime(string, "%Y-%m-%d")


common_parser = ArgumentParser(add_help=False)
common_parser.add_argument(
    "-c",
    "--config",
    type=Path,
    default=DEFAULT_CFG,
    help="Use specific config file [default configs/sequencer.cfg]",
)
common_parser.add_argument(
    "-d",
    "--date",
    help="Date (YYYY-MM-DD) of the start of the night",
    type=valid_date,
)
common_parser.add_argument(
    "-s",
    "--simulate",
    action="store_true",
    default=False,
    help="Do not run, just simulate what would happen",
)
common_parser.add_argument(
    "-t",
    "--test",
    action="store_true",
    default=False,
    help="Avoid interaction with SLURM",
)
common_parser.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    default=False,
    help="Activate debugging mode",
)
# TODO: add here the tel_id common option


def closer_argparser():
    parser = ArgumentParser(parents=[common_parser])

    parser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        dest="noninteractive",
        default=False,
        help="assume yes to all questions",
    )
    parser.add_argument(
        "--seq",
        action="store",
        type=str,
        dest="seqtoclose",
        help="If you only want to close a certain sequence",
    )
    parser.add_argument(
        "--no-dl2",
        action="store_true",
        default=False,
        help="Do not produce DL2 files (default False)",
    )
    parser.add_argument("tel_id", choices=["ST", "LST1", "LST2"])

    return parser


def closercliparsing():
    # parse the command line
    opts = closer_argparser().parse_args()

    # set global variables
    set_common_globals(opts)
    options.seqtoclose = opts.seqtoclose
    options.no_dl2 = opts.no_dl2
    options.noninteractive = opts.noninteractive

    log.debug(f"the options are {opts}")

    # setting the default date and directory if needed
    options.date = set_default_date_if_needed()
    options.directory = analysis_path(options.tel_id)
    set_prod_ids()


def calibration_pipeline_argparser():
    """Command line parser for the calibration pipeline."""
    parser = ArgumentParser(parents=[common_parser])
    parser.add_argument(
        "--prod-id",
        action="store",
        type=str,
        dest="prod_id",
        help="Set the prod ID to define data directories",
    )
    parser.add_argument("--drs4-pedestal-run", type=int, help="DRS4 pedestal run number")
    parser.add_argument("--pedcal-run", type=int, help="Calibration run number")

    parser.add_argument("tel_id", choices=["LST1"])

    return parser


def calibration_pipeline_cliparsing():
    """
    Set the global variables and parse the command
    line arguments for the calibration pipeline
    """
    opts = calibration_pipeline_argparser().parse_args()

    # set global variables
    options.configfile = opts.config
    options.date = opts.date
    options.verbose = opts.verbose
    options.prod_id = opts.prod_id
    options.tel_id = opts.tel_id
    options.simulate = opts.simulate

    # setting the default date and directory if needed
    options.date = set_default_date_if_needed()
    options.directory = analysis_path(options.tel_id)

    return opts.drs4_pedestal_run, opts.pedcal_run


def data_sequence_argparser():
    parser = ArgumentParser(parents=[common_parser])

    parser.add_argument(
        "--prod-id",
        action="store",
        type=str,
        dest="prod_id",
        help="Set the prod ID to define data directories",
    )
    parser.add_argument(
        "--no-dl2",
        action="store_true",
        default=False,
        help="Do not produce DL2 files (default False)",
    )
    parser.add_argument("--pedcal-file", type=Path, help="Path of the calibration file")
    parser.add_argument("--drs4-pedestal-file", type=Path, help="Path of the DRS4 pedestal file")
    parser.add_argument("--time-calib-file", type=Path, help="Path of the time calibration file")
    parser.add_argument(
        "--systematic-correction-file",
        type=Path,
        help="Path of the systematic correction factor file",
    )
    parser.add_argument(
        "--drive-file", type=Path, help="Path of drive log file with pointing information"
    )
    parser.add_argument(
        "--run-summary",
        type=Path,
        help="Path of run summary file with time reference information",
    )
    parser.add_argument(
        "--pedestal-ids-file",
        type=Path,
        help="Path to a file containing the ids of the interleaved pedestal events",
    )
    parser.add_argument("run_number", help="Number of the run to be processed")
    parser.add_argument("tel_id", choices=["ST", "LST1", "LST2"])
    return parser


def data_sequence_cli_parsing():

    # parse the command line
    opts = data_sequence_argparser().parse_args()

    # set global variables
    options.configfile = opts.config.resolve()
    options.date = opts.date
    options.verbose = opts.verbose
    options.simulate = opts.simulate
    options.prod_id = opts.prod_id
    options.no_dl2 = opts.no_dl2
    options.tel_id = opts.tel_id

    log.debug(f"The options and arguments are {opts}")

    # setting the default date and directory if needed
    options.date = set_default_date_if_needed()
    options.directory = analysis_path(options.tel_id)

    set_prod_ids()

    return (
        opts.pedcal_file,
        opts.drs4_pedestal_file,
        opts.time_calib_file,
        opts.systematic_correction_file,
        opts.drive_file,
        opts.run_summary,
        opts.pedestal_ids_file,
        opts.run_number,
    )


def sequencer_argparser():
    """Argument parser for sequencer script."""
    parser = ArgumentParser(
        description="Build the jobs for each run and process them for a given date",
        parents=[common_parser],
    )
    parser.add_argument(
        "--no-submit",
        action="store_true",
        default=False,
        help="Produce job files but do not submit them",
    )
    parser.add_argument(
        "--no-calib",
        action="store_true",
        default=False,
        help="Skip calibration sequence. Run data sequences assuming "
        "calibration products already produced (default False)",
    )
    parser.add_argument(
        "--no-dl2",
        action="store_true",
        default=False,
        help="Do not produce DL2 files (default False)",
    )
    parser.add_argument(
        "tel_id",
        choices=["ST", "LST1", "LST2", "all"],
        help="telescope identifier LST1, LST2, ST or all.",
    )

    return parser


def sequencer_cli_parsing():
    # parse the command line
    opts = sequencer_argparser().parse_args()

    # set global variables
    set_common_globals(opts)
    options.no_submit = opts.no_submit
    options.no_calib = opts.no_calib
    options.no_dl2 = opts.no_dl2

    log.debug(f"the options are {opts}")

    set_prod_ids()

    # setting the default date and directory if needed
    options.date = set_default_date_if_needed()
    options.directory = analysis_path(options.tel_id)


def provprocess_argparser():
    parser = ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=Path,
        default=DEFAULT_CFG,
        help="use specific config file [default configs/sequencer.cfg]",
    )
    parser.add_argument(
        "-f",
        "--filter",
        action="store",
        dest="filter",
        default="",
        help="filter by process granularity [calibration, r0_to_dl1 or dl1_to_dl2]",
    )
    parser.add_argument(
        "-q",
        action="store_true",
        dest="quit",
        help="use this flag to reset session and remove log file",
    )
    parser.add_argument(
        "--no-dl2",
        action="store_true",
        default=False,
        help="Do not produce DL2 files (default False)",
    )
    parser.add_argument(
        "drs4_pedestal_run_id", help="Number of the drs4_pedestal used in the calibration"
    )
    parser.add_argument("pedcal_run_id", help="Number of the used pedcal used in the calibration")
    parser.add_argument("run", help="Number of the run whose provenance is to be extracted")
    parser.add_argument("date", action="store", type=str, help="Observation starting date YYYYMMDD")
    parser.add_argument("prod_id", action="store", type=str, help="Production ID")

    return parser


def provprocessparsing():

    # parse the command line
    opts = provprocess_argparser().parse_args()

    # checking arguments
    if opts.filter not in ["calibration", "r0_to_dl1", "dl1_to_dl2", ""]:
        log.error("incorrect value for --filter argument, type -h for help")

    # set global variables
    options.drs4_pedestal_run_id = opts.drs4_pedestal_run_id
    options.pedcal_run_id = opts.pedcal_run_id
    options.run = opts.run
    options.date = opts.date
    options.configfile = opts.config.resolve()
    options.filter = opts.filter
    options.quit = opts.quit
    options.no_dl2 = opts.no_dl2
    set_prod_ids()


def simproc_argparser():
    parser = ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        default=DEFAULT_CFG,
        help="use specific config file",
    )
    parser.add_argument(
        "-p", action="store_true", dest="provenance", help="produce provenance files"
    )
    parser.add_argument(
        "--force",
        action="store_true",
        dest="force",
        help="force overwrite provenance files",
    )
    parser.add_argument(
        "--append",
        action="store_true",
        dest="append",
        help="append provenance capture to existing prov.log file",
    )
    # parser.add_argument(
    #     "-d",
    #     "--date",
    #     action="store",
    #     type=str,
    #     dest="date",
    #     help="observation ending date YYYY-MM-DD [default today]",
    # )
    # parser.add_argument("tel_id", choices=["ST", "LST1", "LST2"])

    return parser


def simprocparsing():
    opts = simproc_argparser().parse_args()

    # set global variables
    options.prod_id = get_prod_id()
    options.configfile = opts.config
    options.provenance = opts.provenance
    options.force = opts.force
    options.append = opts.append


def copy_datacheck_argparser():
    parser = ArgumentParser(parents=[common_parser])
    parser.add_argument("tel_id", choices=["ST", "LST1", "LST2"])
    return parser


def copy_datacheck_parsing():

    # parse the command line
    opts = copy_datacheck_argparser().parse_args()

    # set global variables
    options.date = opts.date
    options.tel_id = opts.tel_id
    options.configfile = opts.config
    options.date = set_default_date_if_needed()
    options.directory = analysis_path(options.tel_id)
    options.prod_id = get_prod_id()

    if cfg.get("LST1", "DL1_PROD_ID") is not None:
        options.dl1_prod_id = get_dl1_prod_id()
    else:
        options.dl1_prod_id = options.prod_id


def sequencer_webmaker_argparser():
    parser = ArgumentParser(
        description="Script to make an xhtml from LSTOSA sequencer output", parents=[common_parser]
    )
    options.tel_id = "LST1"
    options.prod_id = get_prod_id()

    return parser


def set_default_date_if_needed():
    """Check if the date is set, if not set it to yesterday."""
    return options.date if is_defined(options.date) else YESTERDAY


def set_common_globals(opts):
    """Define common global variables using options module."""
    options.configfile = opts.config.resolve()
    options.date = opts.date
    options.simulate = opts.simulate
    options.test = opts.test
    options.verbose = opts.verbose
    options.tel_id = opts.tel_id


def autocloser_cli_parser():
    """Define the command line parser for the autocloser."""
    parser = ArgumentParser(
        description="Automatic job completion check and sequence closer.", parents=[common_parser]
    )
    parser.add_argument("--ignore-cronlock", action="store_true", help='Ignore "cron.lock"')
    parser.add_argument(
        "-f", "--force", action="store_true", help="Force the autocloser to close the day"
    )
    parser.add_argument(
        "--no-dl2",
        action="store_true",
        default=False,
        help="Disregard the production of DL2 files",
    )
    parser.add_argument("-r", "--runwise", action="store_true", help="Close the day run-wise.")
    parser.add_argument("-l", "--log", type=Path, default=None, help="Write log to a file.")
    parser.add_argument("tel_id", type=str, choices=["LST1"])
    return parser



"""Command line arguments shared variables across all modules."""

configfile = None
stdout = None
stderr = None
date = None
tel_id = None
nightsummary = True
reason = None
directory = None
log_directory = None
simulate = None
noninteractive = None
verbose = None
warning = None
nocheck = None
no_dl2 = None
prod_id = None
dl1_prod_id = None
dl2_prod_id = None
append = None
provenance = True
force = None
filter = None
quit = None
test = False
no_calib = False
no_submit = False
seqtoclose = None
run = None
filters = None



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import datetime
from importlib.metadata import metadata

import osa


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

suppress_warnings = ["ref.citation"]  # ignore citation not referenced warnings

# -- Project information -----------------------------------------------------

project = "lstosa"
author = "lstosa developers"
copyright = f"{author}.  Last updated {datetime.datetime.now().strftime('%d %b %Y %H:%M')}"

version = osa.__version__
# The full version, including alpha/beta/rc tags
release = osa.__version__
python_requires = metadata('lstosa')["Requires-Python"]

# make some variables available to each page (from ctapipe conf.py)
rst_epilog = f"""
.. |python_requires| replace:: {python_requires}
"""

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx_automodapi.automodapi",
    "sphinxarg.ext",
    "numpydoc"
]

numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']
html_static_path = ["_static"]

html_context = {
    "default_mode": "light",
    "github_user": "cta-observatory",
    "github_repo": "lstosa",
    "github_version": "main",
    "doc_path": "docs",
}

html_title = project

# Output file base name for HTML help builder.
htmlhelp_basename = f"{project}doc"

# Refer figures
numfig = True

html_theme_options = {
    "github_url": "https://github.com/cta-observatory/lstosa",
    "navbar_end": ["navbar-icon-links"],
    "icon_links_label": "Quick Links",
    "use_edit_page_button": True,
}



"""
Mock test data set for testing OSA.

R0 files are in BASE_DIR/R0/YYYYMMDD

Calibration files follow the structure:
BASE_DIR/monitoring/PixelCalibration/Cat-A/<calibration_product>/YYYYMMDD/<version>
Where <calibration_product> can be: drs4_baseline, calibration,
drs4_time_sampling_from_FF, ffactor_systematics
<version> will usually be "pro" (for production)

RunSummary file: BASE_DIR/monitoring/RunSummary/RunSummary_YYYMMDD.ecsv

Analysis products (dl1, dl2, muons and datacheck) are produced in the analysis directory:
BASE_DIR/running_analysis/YYYYMMDD/<prod_id>
"""

from textwrap import dedent

import pytest

from osa.configs import options
from osa.configs.config import cfg
from osa.nightsummary.extract import extract_runs, extract_sequences
from osa.nightsummary.nightsummary import run_summary_table
from osa.scripts.tests.test_osa_scripts import run_program
from osa.utils.utils import date_to_dir
from datetime import datetime
import lstchain

date = datetime.fromisoformat("2020-01-17")
nightdir = date_to_dir(date)
prod_id = "v0.1.0"
dl1_prod_id = cfg.get("LST1", "DL1_PROD_ID")
dl2_prod_id = cfg.get("LST1", "DL2_PROD_ID")


@pytest.fixture(scope="session")
def base_test_dir(tmp_path_factory):
    """Creates a temporary directory for the tests."""
    return tmp_path_factory.mktemp("test_files")


@pytest.fixture(scope="session")
def monitoring_dir(base_test_dir):
    monitoring_dir = base_test_dir / "monitoring"
    monitoring_dir.mkdir(parents=True, exist_ok=True)
    return monitoring_dir


@pytest.fixture(scope="session")
def run_summary_dir(monitoring_dir):
    summary_dir = monitoring_dir / "RunSummary"
    summary_dir.mkdir(parents=True, exist_ok=True)
    return summary_dir


@pytest.fixture(scope="session")
def run_catalog_dir(monitoring_dir):
    catalog_dir = monitoring_dir / "RunCatalog"
    catalog_dir.mkdir(parents=True, exist_ok=True)
    return catalog_dir


@pytest.fixture(scope="session")
def calibration_base_dir(monitoring_dir):
    base_dir = monitoring_dir / "PixelCalibration" / "Cat-A"
    base_dir.mkdir(parents=True, exist_ok=True)
    return base_dir


@pytest.fixture(scope="session")
def drive_log(monitoring_dir):
    drive_dir = monitoring_dir / "DrivePositioning"
    drive_file = drive_dir / "DrivePosition_log_20200117.txt"
    drive_dir.mkdir(parents=True, exist_ok=True)
    drive_file.touch()
    return drive_file


@pytest.fixture(scope="session")
def calibration_dir(calibration_base_dir):
    directory = calibration_base_dir / "calibration" / nightdir / f"v{lstchain.__version__}"
    directory.mkdir(parents=True, exist_ok=True)
    return directory


@pytest.fixture(scope="session")
def drs4_baseline_dir(calibration_base_dir):
    directory = calibration_base_dir / "drs4_baseline" / nightdir / f"v{lstchain.__version__}"
    directory.mkdir(parents=True, exist_ok=True)
    return directory


@pytest.fixture(scope="session")
def drs4_time_calibration_files(calibration_base_dir):
    directory = calibration_base_dir / "drs4_time_sampling_from_FF"
    directory1 = directory / "20191124" / "pro"
    directory2 = directory / "20210321" / "pro"
    directory3 = directory / "20210902" / "pro"
    directory1.mkdir(parents=True, exist_ok=True)
    directory2.mkdir(parents=True, exist_ok=True)
    directory3.mkdir(parents=True, exist_ok=True)
    file1 = directory1 / "time_calibration.Run01625.0000.h5"
    file2 = directory2 / "time_calibration.Run04211.0000.h5"
    file3 = directory3 / "time_calibration.Run05979.0000.h5"
    time_file_list = [file1, file2, file3]
    for file in time_file_list:
        file.touch()
    return time_file_list


@pytest.fixture(scope="session")
def systematic_correction_files(calibration_base_dir):
    directory = calibration_base_dir / "ffactor_systematics"
    directory1 = directory / "20200725" / "pro"
    directory2 = directory / "20201110" / "pro"
    directory1.mkdir(parents=True, exist_ok=True)
    directory2.mkdir(parents=True, exist_ok=True)
    file1 = directory1 / "ffactor_systematics_20200725.h5"
    file2 = directory2 / "ffactor_systematics_20201110.h5"
    sys_corr_file_list = [file1, file2]
    for file in sys_corr_file_list:
        file.touch()
    return sys_corr_file_list


@pytest.fixture(scope="session")
def running_analysis_dir(base_test_dir):
    analysis_dir = base_test_dir / "running_analysis" / nightdir / prod_id
    analysis_dir.mkdir(parents=True, exist_ok=True)
    return analysis_dir


@pytest.fixture(scope="session")
def r0_dir(base_test_dir):
    r0_directory = base_test_dir / "R0" / nightdir
    r0_directory.mkdir(parents=True, exist_ok=True)
    return r0_directory


@pytest.fixture(scope="session")
def r0_data(r0_dir):
    r0_files = []
    for i in range(4, 8):
        r0_file = r0_dir / f"LST-1.1.Run0180{i}.0000.fits.fz"
        r0_file.touch()
        r0_files.append(r0_file)
    return r0_files


@pytest.fixture(scope="session")
def dl1b_subdir(running_analysis_dir):
    dl1ab_directory = running_analysis_dir / dl1_prod_id
    dl1ab_directory.mkdir(parents=True, exist_ok=True)
    return dl1ab_directory


@pytest.fixture(scope="session")
def dl2_subdir(running_analysis_dir):
    dl2_directory = running_analysis_dir / dl2_prod_id
    dl2_directory.mkdir(parents=True, exist_ok=True)
    return dl2_directory


@pytest.fixture(scope="session")
def dl2_final_dir(base_test_dir):
    directory = base_test_dir / "DL2" / "20200117" / prod_id / dl2_prod_id
    directory.mkdir(parents=True, exist_ok=True)
    return directory


@pytest.fixture(scope="session")
def calibration_file(calibration_dir):
    """Mock calibration files for testing."""
    calib_file = calibration_dir / "calibration_filters_52.Run01809.0000.h5"
    calib_file.touch()
    return calib_file


@pytest.fixture(scope="session")
def drs4_baseline_file(drs4_baseline_dir):
    """Mock calibration files for testing."""
    drs4_file = drs4_baseline_dir / "drs4_pedestal.Run01804.0000.h5"
    drs4_file.touch()
    return drs4_file


@pytest.fixture(scope="session")
def test_observed_data(running_analysis_dir, dl1b_subdir, dl2_subdir):
    """Mock observed data files for testing."""
    dl1_file = running_analysis_dir / "dl1_LST-1.Run01808.0011.h5"
    muons_file = running_analysis_dir / "muons_LST-1.Run01808.0011.fits"
    interleaved_file = running_analysis_dir / "interleaved_LST-1.Run01808.0011.h5"
    dl1ab_file = dl1b_subdir / "dl1_LST-1.Run01808.0011.h5"
    datacheck_file = dl1b_subdir / "datacheck_dl1_LST-1.Run01808.0011.h5"
    dl2_file = dl2_subdir / "dl2_LST-1.Run01808.0011.h5"
    dl1_file.touch()
    muons_file.touch()
    interleaved_file.touch()
    dl1ab_file.touch()
    datacheck_file.touch()
    dl2_file.touch()
    return dl1_file, dl1ab_file, dl2_file, muons_file, datacheck_file, interleaved_file


@pytest.fixture(scope="session")
def dl2_merged(dl2_final_dir):
    file_1 = dl2_final_dir / "dl2_LST-1.Run01807.h5"
    file_2 = dl2_final_dir / "dl2_LST-1.Run01808.h5"
    file_1.touch()
    file_2.touch()

    return file_1, file_2


@pytest.fixture(scope="session")
def run_summary_file(run_summary_dir):

    summary_content = dedent(
        """\
    # %ECSV 0.9
    # ---
    # datatype:
    # - {name: run_id, datatype: int64}
    # - {name: n_subruns, datatype: int64}
    # - {name: run_type, datatype: string}
    # - {name: ucts_timestamp, datatype: int64}
    # - {name: run_start, datatype: int64}
    # - {name: dragon_reference_time, datatype: int64}
    # - {name: dragon_reference_module_id, datatype: int16}
    # - {name: dragon_reference_module_index, datatype: int16}
    # - {name: dragon_reference_counter, datatype: uint64}
    # - {name: dragon_reference_source, datatype: string}
    # delimiter: ','
    # meta: !!omap
    # - {date: '2020-01-17'}
    # - {lstchain_version: 0.7.0}
    # schema: astropy-2.0
    run_id,n_subruns,run_type,ucts_timestamp,run_start,dragon_reference_time,dragon_reference_module_id,dragon_reference_module_index,dragon_reference_counter,dragon_reference_source
    1803,6,DRS4,1579289727863850890,1579289712000000000,1579289727863850890,90,0,5863850400,ucts
    1804,6,DRS4,1579289727863850890,1579289712000000000,1579289727863850890,90,0,5863850400,ucts
    1805,5,PEDCALIB,1579291426030146503,1579291413000000000,1579291426030146503,90,0,2030146000,ucts
    1806,5,PEDCALIB,1579291932080485703,1579291917000000000,1579291932080485703,90,0,5080485200,ucts
    1807,11,DATA,1579292477145904430,1579292461000000000,1579292477145904430,90,0,6145904000,ucts
    1808,9,DATA,1579292985532016507,1579292975000000000,1579292985532016507,90,0,2532016000,ucts
    1809,5,PEDCALIB,1579291932080485703,1579291917000000000,1579291932080485703,90,0,5080485200,ucts"""
    )

    summary_file = run_summary_dir / "RunSummary_20200117.ecsv"
    summary_file.touch()
    summary_file.write_text(summary_content)
    return summary_file


@pytest.fixture(scope="session")
def run_summary_file_no_calib(run_summary_dir):

    summary_content = dedent(
        """\
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: run_id, datatype: int64}
    # - {name: n_subruns, datatype: int64}
    # - {name: run_type, datatype: string}
    # - {name: ucts_timestamp, datatype: int64}
    # - {name: run_start, datatype: int64}
    # - {name: dragon_reference_time, datatype: int64}
    # - {name: dragon_reference_module_id, datatype: int16}
    # - {name: dragon_reference_module_index, datatype: int16}
    # - {name: dragon_reference_counter, datatype: uint64}
    # - {name: dragon_reference_source, datatype: string}
    # delimiter: ','
    # meta: !!omap
    # - {date: '2022-09-23'}
    # - {lstchain_version: 0.9.6}
    # schema: astropy-2.0
    run_id,n_subruns,run_type,ucts_timestamp,run_start,dragon_reference_time,dragon_reference_module_id,dragon_reference_module_index,dragon_reference_counter,dragon_reference_source
    9379,21,DATA,-1,1663951379000000000,1663951379000000000,0,0,6463544800,run_start
    9380,44,DRS4,1663951929600198168,1663951917000000000,1663951929600198168,0,0,5600197900,ucts
    9381,41,DRS4,1663952467501824402,1663952457000000000,1663952467501824402,0,0,3501824100,ucts"""
    )

    summary_file = run_summary_dir / "RunSummary_20220923.ecsv"
    summary_file.touch()
    summary_file.write_text(summary_content)
    return summary_file


@pytest.fixture(scope="session")
def run_summary_file_no_calib2(run_summary_dir):

    summary_content = dedent(
        """\
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: run_id, datatype: int64}
    # - {name: n_subruns, datatype: int64}
    # - {name: run_type, datatype: string}
    # - {name: ucts_timestamp, datatype: int64}
    # - {name: run_start, datatype: int64}
    # - {name: dragon_reference_time, datatype: int64}
    # - {name: dragon_reference_module_id, datatype: int16}
    # - {name: dragon_reference_module_index, datatype: int16}
    # - {name: dragon_reference_counter, datatype: uint64}
    # - {name: dragon_reference_source, datatype: string}
    # delimiter: ','
    # meta: !!omap
    # - {date: '2022-09-22'}
    # - {lstchain_version: 0.9.6}
    # schema: astropy-2.0
    run_id,n_subruns,run_type,ucts_timestamp,run_start,dragon_reference_time,dragon_reference_module_id,dragon_reference_module_index,dragon_reference_counter,dragon_reference_source
    9326,4,DRS4,1663877740307617642,1663877725000000000,1663877740307617642,0,0,6307617400,ucts
    9327,3,ERRPEDCALIB,1663878027873304499,1663878015000000000,1663878027873304499,0,0,3873304200,ucts
    9258,5,PEDCALIB,1663796417227478675,1663796405000000000,1663796417227478675,90,0,5227478400,ucts"""
    )

    summary_file = run_summary_dir / "RunSummary_20220922.ecsv"
    summary_file.touch()
    summary_file.write_text(summary_content)
    return summary_file


@pytest.fixture(scope="session")
def merged_run_summary(base_test_dir):
    """Mock merged run summary file for testing."""
    summary_content = dedent(
        """\
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: date, datatype: string}
    # - {name: run_id, datatype: int64}
    # - {name: run_type, datatype: string}
    # - {name: n_subruns, datatype: int64}
    # - {name: run_start, datatype: string}
    # - {name: ra, unit: deg, datatype: float64}
    # - {name: dec, unit: deg, datatype: float64}
    # - {name: alt, unit: rad, datatype: float64}
    # - {name: az, unit: rad, datatype: float64}
    # meta: !!omap
    # - __serialized_columns__:
    #     run_start:
    #       __class__: astropy.time.core.Time
    #       format: isot
    #       in_subfmt: '*'
    #       out_subfmt: '*'
    #       precision: 3
    #       scale: utc
    #       value: !astropy.table.SerializedColumn {name: run_start}
    # schema: astropy-2.0
    date run_id run_type n_subruns run_start ra dec alt az
    2019-11-23 1611 DRS4 5 2019-11-23T22:14:09.000 22.0 3.1 0.96 2.3
    2019-11-23 1614 PEDCALIB 10 2019-11-23T23:33:59.000 4.6 2.1 1.1 4.5
    2019-11-23 1615 DATA 61 2019-11-23T23:41:13.000 8.1 45.1 1.5 4.6
    2019-11-23 1616 DATA 62 2019-11-24T00:11:52.000 3.2 4.2 0.9 1.6
    2020-01-17 1804 DRS4 35 2020-01-18T00:44:06.000 2.1 42.9 11.3 4.7
    2020-01-17 1805 PEDCALIB 62 2020-01-18T00:11:52.000 13.9 21.9 17.9 1.6
    2020-01-17 1806 PEDCALIB 35 2020-01-18T00:44:06.000 8.6 29.1 45.5 6.9
    2020-01-17 1807 DATA 35 2020-01-18T00:44:06.000 6.6 2.8 70.4 10.1
    2020-01-17 1808 DATA 35 2020-01-18T00:44:06.000 8.6 9.2 60.8 3.2
    2020-01-17 1809 PEDCALIB 4 2020-01-18T00:44:06.000 6.9 4.2 16.8 11.2
    2022-09-22 9326 DRS4 4 2021-09-22T00:44:06.000 6.9 4.2 16.8 11.2
    2022-09-22 9327 ERRPEDCALIB 3 2021-09-22T00:44:06.000 6.9 4.2 16.8 11.2
    2022-09-22 9258 PEDCALIB 5 2021-09-22T00:44:06.000 6.9 4.2 16.8 11.2
    2022-09-23 9379 DATA 21 2021-09-23T00:44:06.000 6.9 4.2 16.8 11.2
    2022-09-23 9380 DRS4 44 2021-09-23T00:44:06.000 6.9 4.2 16.8 11.2
    2022-09-23 9381 DRS4 41 2021-09-23T00:44:06.000 6.9 4.2 16.8 11.2"""
    )

    merged_summary_dir = base_test_dir / "OSA/Catalog"
    merged_summary_dir.mkdir(parents=True, exist_ok=True)

    file = merged_summary_dir / "merged_RunSummary.ecsv"
    file.touch()
    file.write_text(summary_content)
    return file


@pytest.fixture(scope="session")
def run_summary(run_summary_file):
    """Creates a sequence list from a run summary file."""
    assert run_summary_file.exists()
    return run_summary_table(date)


@pytest.fixture(scope="session")
def pedestal_ids_file(base_test_dir):
    """Mock pedestal ids file for testing."""
    pedestal_ids_dir = base_test_dir / "auxiliary/PedestalFinder/20200117"
    pedestal_ids_dir.mkdir(parents=True, exist_ok=True)
    file = pedestal_ids_dir / "pedestal_ids_Run01808.0000.h5"
    file.touch()
    return file


@pytest.fixture(scope="session")
def sequence_list(
    running_analysis_dir,
    run_summary,
    drs4_time_calibration_files,
    systematic_correction_files,
    r0_data,
    pedestal_ids_file,
    merged_run_summary,
):
    """Creates a sequence list from a run summary file."""
    options.directory = running_analysis_dir
    options.simulate = True
    options.test = True

    for file in drs4_time_calibration_files:
        assert file.exists()

    for file in systematic_correction_files:
        assert file.exists()

    for file in r0_data:
        assert file.exists()

    assert pedestal_ids_file.exists()
    assert merged_run_summary.exists()

    run_list = extract_runs(run_summary)
    options.test = False
    return extract_sequences(options.date, run_list)


@pytest.fixture(scope="session")
def sequence_file_list(
    running_analysis_dir,
    run_summary_file,
    run_catalog,
    drs4_time_calibration_files,
    systematic_correction_files,
    r0_data,
):
    for r0_file in r0_data:
        assert r0_file.exists()

    for file in drs4_time_calibration_files:
        assert file.exists()

    for file in systematic_correction_files:
        assert file.exists()

    assert run_summary_file.exists()
    assert run_catalog.exists()

    run_program("sequencer", "-d", "2020-01-17", "--no-submit", "-t", "LST1")
    # First sequence in the list corresponds to the calibration run 1809
    return [
        running_analysis_dir / "sequence_LST1_01809.py",
        running_analysis_dir / "sequence_LST1_01807.py",
        running_analysis_dir / "sequence_LST1_01808.py",
    ]


@pytest.fixture(scope="session")
def txt_file_test(running_analysis_dir):
    from osa.utils.iofile import write_to_file

    options.simulate = False
    file = running_analysis_dir / "test.txt"
    write_to_file(file, "This is a test")
    options.simulate = True
    return file


@pytest.fixture(scope="session")
def datacheck_dl1_files(base_test_dir):
    dl1b_dir = base_test_dir / "DL1" / "20200117" / "v0.1.0" / "tailcut84"
    dl1b_dir.mkdir(parents=True, exist_ok=True)
    pdf_file_1 = dl1b_dir / "datacheck_dl1_LST-1.Run01807.pdf"
    pdf_file_2 = dl1b_dir / "datacheck_dl1_LST-1.Run01808.pdf"
    pdf_file_1.touch()
    pdf_file_2.touch()
    return pdf_file_1, pdf_file_2


@pytest.fixture(scope="session")
def longterm_dir(base_test_dir):
    directory = base_test_dir / "OSA" / "DL1DataCheck_LongTerm" / "v0.1.0" / "20200117"
    directory.mkdir(parents=True, exist_ok=True)
    return directory


@pytest.fixture(scope="session")
def daily_datacheck_dl1_files(longterm_dir):
    html_file = longterm_dir / "DL1_datacheck_20200117.html"
    h5_file = longterm_dir / "DL1_datacheck_20200117.h5"
    log_file = longterm_dir / "DL1_datacheck_20200117.log"
    html_file.touch()
    h5_file.touch()
    log_file.touch()
    return html_file, h5_file, log_file


@pytest.fixture(scope="session")
def calibration_check_plot(calibration_dir):
    calibration_dir_log = calibration_dir / "log"
    calibration_dir_log.mkdir(parents=True, exist_ok=True)
    file = calibration_dir_log / "calibration_filters_52.Run01809.0000.pdf"
    file.touch()
    return file


@pytest.fixture(scope="session")
def drs4_check_plot(drs4_baseline_dir):
    drs4_baseline_dir_log = drs4_baseline_dir / "log"
    drs4_baseline_dir_log.mkdir(parents=True, exist_ok=True)
    file = drs4_baseline_dir_log / "drs4_pedestal.Run01804.0000.pdf"
    file.touch()
    return file


@pytest.fixture(scope="session")
def run_catalog(run_catalog_dir):
    source_information = dedent(
        """\
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: run_id, datatype: int32}
    # - {name: source_name, datatype: string}
    # - {name: source_ra, datatype: float64}
    # - {name: source_dec, datatype: float64}
    # delimiter: ','
    # schema: astropy-2.0
    run_id,source_name,source_ra,source_dec
    1807,Crab,83.543,22.08
    1808,MadeUpSource,115.441,43.98"""
    )

    catalog_file = run_catalog_dir / "RunCatalog_20200117.ecsv"
    catalog_file.touch()
    catalog_file.write_text(source_information)
    return catalog_file


@pytest.fixture(scope="session")
def database(base_test_dir):
    import sqlite3

    osa_dir = base_test_dir / "OSA"
    osa_dir.mkdir(parents=True, exist_ok=True)
    db_file = osa_dir / "osa.db"
    with sqlite3.connect(db_file) as connection:
        cursor = connection.cursor()
        cursor.execute(
            """CREATE TABLE IF NOT EXISTS processing
            (telescope, date, prod_id, start, end, is_finished)"""
        )
        cursor.connection.commit()
        yield cursor



#!/usr/bin/env python
"""
Script to copy analysis products to datacheck webserver creating new
directories whenever they are needed.
"""

import logging

from osa.configs import options
from osa.paths import (
    datacheck_directory,
    get_datacheck_files,
    destination_dir,
)
from osa.utils.cliopts import copy_datacheck_parsing
from osa.utils.logging import myLogger
from osa.utils.utils import DATACHECK_FILE_PATTERNS, date_to_dir
from osa.webserver.utils import copy_to_webserver

log = myLogger(logging.getLogger())


def are_files_copied(data_type: str, files: list) -> bool:
    """Check if all files of a given data type are copied."""
    n_files = len(files)

    if n_files == 0:
        log.warning(f"No {data_type} files found.")
        return False

    if data_type == "PEDESTAL" and n_files != 1:
        log.warning(f"Expected at least 1 PDF file, {n_files} found.")
        return False

    if data_type == "CALIB" and n_files != 1:
        log.warning(f"Expected at least 1 PDF file, {n_files} found.")
        return False

    if data_type == "DL1AB" and n_files != get_number_of_runs():
        log.warning(f"Expected {get_number_of_runs()} PDF files, {n_files} found.")
        return False

    if data_type == "LONGTERM" and n_files != 3:
        log.warning(f"Expected 3 DL1 check files (HTML, h5 and log), {n_files} found.")
        return False

    return True


def main():
    """Copy datacheck products to the webserver."""
    log.setLevel(logging.INFO)

    log.info(
        """
        ___________________________________________________________________________
        Expected PDF datacheck files: DRS4, ENF calibration & run-wise DL1.
        Additionally, daily DL1 log, HTML and h5 files should be copied.
        If any of these are missing it might happen that they are not produced yet.
        ___________________________________________________________________________
        """
    )

    copy_datacheck_parsing()
    nightdir = date_to_dir(options.date)

    all_files_are_copied = False

    for data_type, pattern in DATACHECK_FILE_PATTERNS.items():
        log.info(f"Looking for {pattern}")
        directory = datacheck_directory(data_type=data_type, date=nightdir)
        files = get_datacheck_files(pattern, directory)
        if len(files) != 0:
            copy_to_webserver(files, data_type, nightdir, options.prod_id)

        # Check if all files are copied
        all_files_are_copied = are_files_copied(data_type, files)

    if all_files_are_copied:
        log.info("All datacheck files copied. No more files are expected.")
    else:
        log.warning(
            "Not all datacheck files were copied. Check for problems or whether "
            "data were actually taken."
        )


def get_number_of_runs():
    """
    Get the run sequence processed list for the given date by globbing the
    run-wise DL1 files.
    """
    dl1_directory = destination_dir("DL1AB", create_dir=False)
    list_files = list(dl1_directory.glob("dl1_LST-1.Run?????.h5"))
    return len(list_files)


if __name__ == "__main__":
    main()



"""Provenance capture functions."""

import datetime
import hashlib
import logging
import logging.config
import os
import platform
import sys
import uuid
from functools import wraps
from pathlib import Path

import pkg_resources
import psutil
import yaml

from osa.provenance.utils import get_log_config, parse_variables

# gammapy specific
# from gammapy.scripts.info import (
#    get_info_dependencies,
#    get_info_envvar,
#    get_info_version,
# )

__all__ = ["trace", "get_file_hash", "get_activity_id"]

from osa.utils.logging import myLogger

_interesting_env_vars = [
    "CONDA_DEFAULT_ENV",
    "CONDA_PREFIX",
    "CONDA_PYTHON_EXE",
    "CONDA_EXE",
    "CONDA_PROMPT_MODIFIER",
    "CONDA_SHLVL",
    "PATH",
    "LD_LIBRARY_PATH",
    "DYLD_LIBRARY_PATH",
    "USER",
    "HOME",
    "SHELL",
]

CONFIG_PATH = Path(__file__).resolve().parent / "config"
SCHEMA_FILE = CONFIG_PATH / "definition.yaml"
# LOGGER_FILE = CONFIG_PATH / "logger.yaml"
definition = yaml.safe_load(SCHEMA_FILE.read_text())
provconfig = yaml.safe_load(get_log_config())
logger = logging.getLogger("provLogger")
LOG_FILENAME = provconfig["handlers"]["provHandler"]["filename"]
PROV_PREFIX = provconfig["PREFIX"]
SUPPORTED_HASH_METHOD = ["md5"]
SUPPORTED_HASH_BUFFER = ["content", "path"]
REDUCTION_TASKS = ["r0_to_dl1", "dl1ab", "dl1_datacheck", "dl1_to_dl2"]

# global variables
traced_entities = {}
session_name = ""
session_tag = ""


def setup_logging():
    """Setup logging configuration."""
    log = myLogger(logging.getLogger(__name__))

    try:
        logging.config.dictConfig(provconfig)
    except Exception as ex:
        log.warning(ex)
        log.warning("Failed to set up the logger.")


# def provenance(cls):
#     """A function decorator which decorates the methods of a class with trace function."""
#
#     setup_logging()
#     for attr in cls.__dict__:
#         if attr in definition["activities"].keys() and callable(getattr(cls, attr)):
#             setattr(cls, attr, trace(getattr(cls, attr)))
#     return cls


def trace(func):
    """Trace and capture provenance info inside a method /function."""
    setup_logging()

    @wraps(func)
    def wrapper(*args, **kwargs):

        activity = func.__name__
        activity_id = get_activity_id()
        # class_instance = args[0]
        class_instance = func
        class_instance.args = args
        class_instance.kwargs = kwargs

        # OSA specific
        # variables parsing
        global session_name, session_tag
        class_instance = parse_variables(class_instance)
        if class_instance.__name__ in REDUCTION_TASKS:
            session_tag = f"{activity}:{class_instance.ObservationRun}"
            session_name = f"{class_instance.ObservationRun}"
        else:
            session_tag = f"{activity}:{class_instance.PedestalRun}-{class_instance.CalibrationRun}"
            session_name = f"{class_instance.PedestalRun}-{class_instance.CalibrationRun}"
        # OSA specific
        # variables parsing

        # provenance capture before execution
        derivation_records = get_derivation_records(class_instance, activity)
        parameter_records = get_parameters_records(class_instance, activity, activity_id)
        usage_records = get_usage_records(class_instance, activity, activity_id)

        # activity execution
        start = datetime.datetime.now().isoformat()
        result = func(*args, **kwargs)
        end = datetime.datetime.now().isoformat()

        # no provenance logging
        if not log_is_active(class_instance, activity):
            return result
        # provenance logging only if activity ends properly
        session_id = log_session(class_instance, start)
        for log_record in derivation_records:
            log_prov_info(log_record)
        log_start_activity(activity, activity_id, session_id, start)
        for log_record in parameter_records:
            log_prov_info(log_record)
        for log_record in usage_records:
            log_prov_info(log_record)
        log_generation(class_instance, activity, activity_id)
        log_finish_activity(activity_id, end)
        return result

    return wrapper


def log_is_active(class_instance, activity):
    """Check if provenance option is enabled in configuration settings."""
    active = True
    if activity not in definition["activities"].keys():
        active = False
    if not class_instance:
        active = False
    if "capture" not in provconfig or not provconfig["capture"]:
        active = False
    return active


def get_activity_id():
    # uuid example: ea5caa9f-0a76-42f5-a1a7-43752df755f0
    # uuid[-12:]: 43752df755f0
    # uuid[-6:]: f755f0
    return str(uuid.uuid4())[-6:]


def get_hash_method():
    """Helper function that returns hash method used."""
    try:
        method = provconfig["HASH_METHOD"]
    except KeyError:
        method = "md5"
    if method not in SUPPORTED_HASH_METHOD:
        logger.warning(f"Hash method {method} not supported.")
        method = "md5"
    return method


def get_hash_buffer():
    """Helper function that returns buffer content to be used in hash method used."""
    try:
        buffer = provconfig["HASH_TYPE"]
    except KeyError:
        buffer = "path"
    if buffer not in SUPPORTED_HASH_BUFFER:
        logger.warning(f"Hash buffer {buffer} not supported.")
        buffer = "path"
    return buffer


def get_file_hash(str_path, buffer=get_hash_buffer(), method=get_hash_method()):
    """Helper function that returns hash of the content of a file."""
    file_hash = ""
    full_path = Path(str_path)
    hash_func = getattr(hashlib, method)()
    if buffer == "content":
        if full_path.is_file():
            with open(full_path, "rb") as f:
                block_size = 65536
                buf = f.read(block_size)
                while len(buf) > 0:
                    hash_func.update(buf)
                    buf = f.read(block_size)
            file_hash = hash_func.hexdigest()
            logger.debug(f"File entity {str_path} has {method} hash {file_hash}")
            return file_hash

        logger.warning(f"File entity {str_path} not found")
        return str_path

    if not file_hash:
        hash_func.update(str(full_path).encode())
        return hash_func.hexdigest()


def get_entity_id(value, item):
    """Helper function that makes the id of an entity, depending on its type."""
    try:
        entity_name = item["entityName"]
        entity_type = definition["entities"][entity_name]["type"]
    except Exception as ex:
        logger.warning(f"Not found in model {ex} in {item}")
        entity_name = ""
        entity_type = ""

    # gammapy specific
    # if entity_type == "FileCollection":
    #     filename = value
    #     index = definition["entities"][entity_name].get("index", "")
    #     if Path(filename).is_dir() and index:
    #         filename = Path(value) / index
    #     return get_file_hash(filename)
    if entity_type == "File":
        # osa specific hash path
        # async calls does not allow for hash content
        return get_file_hash(value, buffer="path")
        # osa specific hash path
        # async calls does not allow for hash content
    try:
        entity_id = abs(hash(value) + hash(str(value)))
        if hasattr(value, "entity_version"):
            entity_id += getattr(value, "entity_version")
        return entity_id
    except TypeError:
        # remark: two different objects may use the same memory address
        # so use hash(entity_name) to avoid issues
        return abs(id(value) + hash(entity_name))


def get_nested_value(nested, branch):
    """Helper function that gets a specific value in a nested dictionary or class."""
    list_branch = branch.split(".")
    leaf = list_branch.pop(0)
    # return value of leaf
    if not nested:
        return globals().get(leaf, None)
    # get value of leaf
    if isinstance(nested, dict):
        val = nested.get(leaf, None)
    elif isinstance(nested, object):
        if "(" in leaf:  # leaf is a function
            leaf_elements = leaf.replace(")", "").replace(" ", "").split("(")
            leaf_arg_list = leaf_elements.pop().split(",")
            leaf_func = leaf_elements.pop()
            leaf_args = []
            leaf_kwargs = {}
            for arg in leaf_arg_list:
                if "=" in arg:
                    k, v = arg.split("=")
                    leaf_kwargs[k] = v.replace('"', "")
                elif arg:
                    leaf_args.append(arg.replace('"', ""))
            val = getattr(nested, leaf_func, lambda *args, **kwargs: None)(
                *leaf_args, **leaf_kwargs
            )
        else:  # leaf is an attribute
            val = getattr(nested, leaf, None)
    else:
        raise TypeError
    # continue to explore branch
    if len(list_branch):
        str_branch = ".".join(list_branch)
        return get_nested_value(val, str_branch)
    # return value of leaf
    if not val:
        val = globals().get(leaf, None)
    return val


def get_item_properties(nested, item):
    """Helper function that returns properties of an entity or member."""
    try:
        entity_name = item["entityName"]
        entity_type = definition["entities"][entity_name]["type"]
    except Exception as ex:
        logger.warning(f"{ex} in {item}")
        entity_name = ""
        entity_type = ""
    properties = {}
    if "id" in item:
        item_id = str(get_nested_value(nested, item["id"]))
        item_ns = item.get("namespace", None)
        if item_ns:
            item_id = item_ns + ":" + item_id
        properties["id"] = item_id
    if "location" in item:
        properties["location"] = get_nested_value(nested, item["location"])
    value = get_nested_value(nested, item["value"]) if "value" in item else ""
    if not value and "location" in properties:
        value = properties["location"]
    if "overwrite" in item:
        # add or increment entity_version to make value a different entity
        if hasattr(value, "entity_version"):
            version = getattr(value, "entity_version")
            version += 1
            setattr(value, "entity_version", version)
        else:
            try:
                setattr(value, "entity_version", 1)
            except AttributeError as ex:
                logger.warning(f"{ex} for {value}")
    if value and "id" not in properties:
        properties["id"] = get_entity_id(value, item)
        if "File" in entity_type:
            properties["filepath"] = value
            if properties["id"] != value:
                method = get_hash_method()
                properties["hash"] = properties["id"]
                properties["hash_type"] = method

    if entity_name:
        properties["name"] = entity_name
        for attr in ["type", "contentType"]:
            if attr in definition["entities"][entity_name]:
                properties[attr] = definition["entities"][entity_name][attr]
    return properties


def get_python_packages():
    """Return the collection of dependencies available for importing."""
    return [
        {"name": p.project_name, "version": p.version, "path": p.module_path}
        for p in sorted(pkg_resources.working_set, key=lambda p: p.project_name)
    ]


def log_prov_info(prov_dict):
    """Write a dictionary to the logger."""
    # OSA specific session tag used in merging prov from parallel sessions
    prov_dict["session_tag"] = session_tag
    #
    record_date = datetime.datetime.now().isoformat()
    logger.info(f"{PROV_PREFIX}{record_date}{PROV_PREFIX}{prov_dict}")


def log_session(class_instance, start):
    """Log start of a session."""
    # OSA specific
    # prov session is outside scripting and is run-wise
    # we may have different sessions/runs in the same log file
    # session_id = abs(hash(class_instance))
    if class_instance.__name__ in REDUCTION_TASKS:
        session_id = f"{class_instance.ObservationDate}{class_instance.ObservationRun}"
    else:
        session_id = f"{class_instance.PedestalRun}{class_instance.CalibrationRun}"
    # OSA specific
    # prov session is outside scripting and is run-wise
    # we may have different sessions/runs in the same log file

    system = get_system_provenance()
    log_record = {
        "session_id": session_id,
        "name": session_name,
        "startTime": start,
        "system": system,
        # OSA specific
        "observation_date": class_instance.ObservationDate,
        # OSA specific
        "software_version": class_instance.SoftwareVersion,
        "config_file": class_instance.ProcessingConfigFile,
        "config_file_hash": get_file_hash(class_instance.ProcessingConfigFile, buffer="path"),
        "config_file_hash_type": get_hash_method(),
    }
    if class_instance.__name__ in REDUCTION_TASKS:
        log_record["observation_run"] = class_instance.ObservationRun  # a session is run-wise
    else:
        log_record["pedestal_run"] = class_instance.PedestalRun
        log_record["calibration_run"] = class_instance.CalibrationRun
    log_prov_info(log_record)
    return session_id


def log_start_activity(activity, activity_id, session_id, start):
    """Log start of an activity."""
    log_record = {
        "activity_id": activity_id,
        "name": activity,
        "startTime": start,
        "in_session": session_id,
        "agent_name": os.getenv("USER", "Anonymous"),
        "script": sys.argv[0],
    }
    log_prov_info(log_record)


def log_finish_activity(activity_id, end):
    """Log end of an activity."""
    log_record = {"activity_id": activity_id, "endTime": end}
    log_prov_info(log_record)


def get_derivation_records(class_instance, activity):
    """Get log records for potentially derived entity."""
    records = []
    for var, pair in traced_entities.items():
        entity_id, item = pair
        value = get_nested_value(class_instance, var)
        if value:
            new_id = get_entity_id(value, item)
            if new_id != entity_id:
                log_record = {"entity_id": new_id, "progenitor_id": entity_id}
                records.append(log_record)
                traced_entities[var] = (new_id, item)
                logger.warning(f"Derivation detected in {activity} for {var}. ID: {new_id}")
    return records


def get_parameters_records(class_instance, activity, activity_id):
    """Get log records for parameters of the activity."""
    records = []
    parameter_list = definition["activities"][activity]["parameters"] or []
    if parameter_list:
        parameters = {}
        for parameter in parameter_list:
            if "name" in parameter and "value" in parameter:
                parameter_value = get_nested_value(class_instance, parameter["value"])
                if parameter_value:
                    parameters[parameter["name"]] = parameter_value
        if parameters:
            log_record = {"activity_id": activity_id, "parameters": parameters}
            records.append(log_record)
    return records


def get_usage_records(class_instance, activity, activity_id):
    """Get log records for each usage of the activity."""
    records = []
    usage_list = definition["activities"][activity]["usage"] or []
    for item in usage_list:
        props = get_item_properties(class_instance, item)
        if "id" in props:
            entity_id = props.pop("id")
            # record usage
            log_record = {
                "activity_id": activity_id,
                "used_id": entity_id,
            }
            if "role" in item:
                log_record.update({"used_role": item["role"]})
            # record entity
            log_record_ent = {
                "entity_id": entity_id,
            }
            if "entityName" in item:
                log_record_ent.update({"name": item["entityName"]})
            for prop in props:
                log_record_ent.update({prop: props[prop]})
            records.append(log_record_ent)
            records.append(log_record)
    return records


def log_generation(class_instance, activity, activity_id):
    """Log generated entities."""
    generation_list = definition["activities"][activity]["generation"] or []
    for item in generation_list:
        props = get_item_properties(class_instance, item)
        if "id" in props:
            entity_id = props.pop("id")
            # record generation
            if "value" in item:
                traced_entities[item["value"]] = (entity_id, item)
            log_record = {
                "activity_id": activity_id,
                "generated_id": entity_id,
            }
            if "role" in item:
                log_record.update({"generated_role": item["role"]})
            # record entity
            log_record_ent = {"entity_id": entity_id}
            if "entityName" in item:
                log_record_ent.update({"name": item["entityName"]})
            for prop in props:
                log_record_ent.update({prop: props[prop]})
            log_prov_info(log_record_ent)
            log_prov_info(log_record)
            if "has_members" in item:
                log_members(entity_id, item["has_members"], class_instance)
            if "has_progenitors" in item:
                log_progenitors(entity_id, item["has_progenitors"], class_instance)


def log_members(entity_id, subitem, class_instance):
    """Log members of and entity."""
    if "list" in subitem:
        member_list = get_nested_value(class_instance, subitem["list"]) or []
    else:
        member_list = [class_instance]
    for member in member_list:
        props = get_item_properties(member, subitem)
        if "id" in props:
            mem_id = props.pop("id")
            # record membership
            log_record = {
                "entity_id": entity_id,
                "member_id": mem_id,
            }
            # record entity
            log_record_ent = {"entity_id": mem_id}
            if "entityName" in subitem:
                log_record_ent.update({"name": subitem["entityName"]})
            for prop in props:
                log_record_ent.update({prop: props[prop]})
            log_prov_info(log_record_ent)
            log_prov_info(log_record)


def log_progenitors(entity_id, subitem, class_instance):
    """Log progenitors of and entity."""
    if "list" in subitem:
        progenitor_list = get_nested_value(class_instance, subitem["list"]) or []
    else:
        progenitor_list = [class_instance]
    for entity in progenitor_list:
        props = get_item_properties(entity, subitem)
        if "id" in props:
            progen_id = props.pop("id")
            # record progenitor link
            log_record = {
                "entity_id": entity_id,
                "progenitor_id": progen_id,
            }
            # record entity
            log_record_ent = {"entity_id": progen_id}
            for prop in props:
                log_record_ent.update({prop: props[prop]})
            log_prov_info(log_record_ent)
            log_prov_info(log_record)


# def log_file_generation(str_path, entity_name="", used=None, role="", activity_name=""):
#     """Log properties of a generated file."""
#
#     if used is None:
#         used = []
#     if Path(str_path).isfile():
#         method = get_hash_method()
#         item = dict(
#             file_path=str_path,
#             entityName=entity_name,
#         )
#         entity_id = get_entity_id(str_path, item)
#         log_record = {
#             "entity_id": entity_id,
#             "name": entity_name,
#             "location": str_path,
#             "hash": entity_id,
#             "hash_type": method,
#         }
#         log_prov_info(log_record)
#         if activity_name:
#             activity_id = get_activity_id()
#             log_record = {
#                 "activity_id": activity_id,
#                 "name": activity_name,
#             }
#             log_prov_info(log_record)
#             for used_entity in used:
#                 used_id = get_entity_id(used_entity, {})
#                 log_record = {
#                     "activity_id": activity_id,
#                     "used_id": used_id,
#                 }
#                 log_prov_info(log_record)
#             log_record = {
#                 "activity_id": activity_id,
#                 "generated_id": entity_id,
#             }
#             if role:
#                 log_record.update({"generated_role": role})
#             log_prov_info(log_record)
#         else:
#             for used_entity in used:
#                 used_id = get_entity_id(used_entity, {})
#                 log_record = {
#                     "entity_id": entity_id,
#                     "progenitor_id": used_id,
#                 }
#                 log_prov_info(log_record)


# ctapipe inherited code
#
#
def get_system_provenance():
    """
    Return JSON string containing provenance for all
    things that are fixed during the runtime.
    """

    bits, linkage = platform.architecture()

    return dict(
        # gammapy specific
        # version=get_info_version(),
        # dependencies=get_info_dependencies(),
        # envvars=get_info_envvar(),
        executable=sys.executable,
        platform=dict(
            architecture_bits=bits,
            architecture_linkage=linkage,
            machine=platform.machine(),
            processor=platform.processor(),
            node=platform.node(),
            version=str(platform.version()),
            system=platform.system(),
            release=platform.release(),
            libcver=str(platform.libc_ver()),
            num_cpus=psutil.cpu_count(),
            boot_time=datetime.datetime.fromtimestamp(psutil.boot_time()).isoformat(),
        ),
        python=dict(
            version_string=sys.version,
            version=platform.python_version(),
            compiler=platform.python_compiler(),
            implementation=platform.python_implementation(),
            packages=get_python_packages(),
        ),
        environment=get_env_vars(),
        arguments=sys.argv,
        start_time_utc=datetime.datetime.now().isoformat(),
    )


def get_env_vars():
    """Return env vars defined at the main scope of the script."""
    return {var: os.getenv(var, None) for var in _interesting_env_vars}



#!/usr/bin/env python

"""
Calibration pipeline

Script to process the pedestal and calibration runs to produce the
DRS4 pedestal and charge calibration files. It pipes together the two
onsite calibration scripts from lstchain.
"""

import logging
import sys
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.job import historylevel
from osa.paths import drs4_pedestal_exists, calibration_file_exists
from osa.provenance.capture import trace
from osa.utils.cliopts import calibration_pipeline_cliparsing
from osa.utils.logging import myLogger
from osa.workflow.stages import DRS4PedestalStage, ChargeCalibrationStage

__all__ = [
    "calibration_sequence",
    "calibrate_charge",
    "drs4_pedestal",
    "drs4_pedestal_command",
    "calibration_file_command",
]

log = myLogger(logging.getLogger())


def is_calibration_produced(drs4_pedestal_run_id: int, pedcal_run_id: int) -> bool:
    """
    Check if both daily calibration (DRS4 baseline and
    charge calibration) files are already produced.
    """
    return drs4_pedestal_exists(drs4_pedestal_run_id, options.prod_id) \
        and calibration_file_exists(pedcal_run_id, options.prod_id)


def drs4_pedestal_command(drs4_pedestal_run_id: int) -> list:
    """Build the create_drs4_pedestal command."""
    base_dir = Path(cfg.get("LST1", "BASE")).resolve()
    return [
        "onsite_create_drs4_pedestal_file",
        f"--run_number={drs4_pedestal_run_id}",
        f"--base_dir={base_dir}",
        "--no-progress",
    ]


def calibration_file_command(drs4_pedestal_run_id: int, pedcal_run_id: int) -> list:
    """Build the create_calibration_file command."""
    base_dir = Path(cfg.get("LST1", "BASE")).resolve()
    cmd = [
        "onsite_create_calibration_file",
        f"--pedestal_run={drs4_pedestal_run_id}",
        f"--run_number={pedcal_run_id}",
        f"--base_dir={base_dir}",
    ]
    # In case of problems with trigger tagging:
    if cfg.getboolean("lstchain", "use_ff_heuristic_id"):
        cmd.append("--flatfield-heuristic")
    return cmd


def calibration_sequence(drs4_pedestal_run_id: int, pedcal_run_id: int) -> int:
    """
    Handle the two stages for creating the daily calibration products:
    DRS4 pedestal and charge calibration files.

    Parameters
    ----------
    drs4_pedestal_run_id : int
    pedcal_run_id : int

    Returns
    -------
    rc : int
        Return code
    """
    analysis_dir = Path(options.directory)
    history_file = analysis_dir / f"sequence_LST1_{pedcal_run_id:05d}.history"

    level, rc = (2, 0) if options.simulate else historylevel(history_file, "PEDCALIB")

    log.info(f"Going to level {level}")

    if level == 2:
        rc = drs4_pedestal(drs4_pedestal_run_id, pedcal_run_id)
        level -= 1
        log.info(f"Going to level {level}")
    if level == 1:
        rc = calibrate_charge(drs4_pedestal_run_id, pedcal_run_id)
        level -= 1
        log.info(f"Going to level {level}")
    if level == 0:
        log.info(f"Job for sequence {pedcal_run_id} finished without fatal errors")

    return rc


@trace
def drs4_pedestal(
    drs4_pedestal_run_id: int,
    pedcal_run_id: int,
) -> int:
    """
    Create a DRS4 pedestal file for baseline correction.

    Parameters
    ----------
    drs4_pedestal_run_id : int
        DRS4 pedestal run number
    pedcal_run_id : int
        PEDCALIB run number

    Returns
    -------
    rc : int
        Return code
    """
    if options.simulate or drs4_pedestal_exists(drs4_pedestal_run_id, options.prod_id):
        return 0

    cmd = drs4_pedestal_command(drs4_pedestal_run_id)

    analysis_step = DRS4PedestalStage(
        run=f"{drs4_pedestal_run_id:05d}", run_pedcal=f"{pedcal_run_id:05d}", command_args=cmd
    )
    analysis_step.execute()
    return analysis_step.rc


@trace
def calibrate_charge(
    drs4_pedestal_run_id: int,
    pedcal_run_id: int,
) -> int:
    """
    Create the calibration file to transform from ADC counts to photo-electrons

    Parameters
    ----------
    drs4_pedestal_run_id : int
        String with run number of the pedestal run
    pedcal_run_id : int
        String with run number of the pedcal run

    Returns
    -------
    rc: int
        Return code
    """
    if options.simulate or calibration_file_exists(pedcal_run_id, options.prod_id):
        return 0

    cmd = calibration_file_command(
        drs4_pedestal_run_id=drs4_pedestal_run_id, pedcal_run_id=pedcal_run_id
    )
    analysis_step = ChargeCalibrationStage(run=f"{pedcal_run_id:05d}", command_args=cmd)

    try:
        analysis_step.execute()
        return analysis_step.rc

    except Exception:
        log.info(f"Failed. Return code {analysis_step.rc}")
        cmd.append("--filters=52")
        log.info("Trying again by setting filters 52")
        analysis_step = ChargeCalibrationStage(run=f"{pedcal_run_id:05d}", command_args=cmd)
        analysis_step.execute()
        return analysis_step.rc


def main():
    """
    Performs the calibration steps (obtain the drs4 baseline correction
    and ADC to photo-electron coefficients)
    """
    drs4_pedestal_run, pedcal_run = calibration_pipeline_cliparsing()

    if options.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    if is_calibration_produced(drs4_pedestal_run_id=drs4_pedestal_run, pedcal_run_id=pedcal_run):
        log.info(
            f"Calibration files already produced from "
            f"runs {drs4_pedestal_run:05d} and {pedcal_run:05d}"
        )
        sys.exit(0)

    rc = calibration_sequence(drs4_pedestal_run, pedcal_run)
    sys.exit(rc)


if __name__ == "__main__":
    main()



"""
Functions to compute significance and build the theta2 plots for each source.

Based on Explore_DL2_data.ipynb from LST-1 analysis 2022 school.
"""

import logging
import subprocess
from datetime import datetime
from pathlib import Path
from typing import MutableMapping, Any

import astropy
import astropy.units as u
import click
import numpy as np
import pandas as pd
import toml
from ctapipe.containers import EventType
from gammapy.stats import WStatCountsStatistic
from lstchain.io.io import dl2_params_lstcam_key
from lstchain.reco.utils import get_effective_time, extract_source_position, compute_theta2
from matplotlib import pyplot as plt

from osa.configs import options
from osa.configs.config import cfg
from osa.nightsummary.extract import get_source_list
from osa.paths import DEFAULT_CFG, destination_dir, analysis_path
from osa.utils.cliopts import get_prod_id, get_dl2_prod_id
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_dir, YESTERDAY

__all__ = [
    "create_hist",
    "lima_significance",
    "event_selection",
    "plot_theta2",
]

from osa.webserver.utils import directory_in_webserver

log = myLogger(logging.getLogger(__name__))

SELECTION_CUTS_FILE = Path(__file__).parent / "selection_cuts.toml"

mpl_linewidth = 1.6
mpl_rc = {
    "figure.autolayout": True,
    "figure.dpi": 150,
    "font.size": 12,
    "lines.linewidth": mpl_linewidth,
    "axes.grid": True,
    "grid.linestyle": ":",
    "grid.linewidth": 1.4,
    "axes.linewidth": mpl_linewidth,
    "xtick.major.size": 7,
    "xtick.major.width": mpl_linewidth,
    "xtick.minor.size": 4,
    "xtick.minor.width": mpl_linewidth,
    "xtick.minor.visible": False,
    "ytick.major.size": 7,
    "ytick.major.width": mpl_linewidth,
    "ytick.minor.size": 4,
    "ytick.minor.width": mpl_linewidth,
    "ytick.minor.visible": False,
}
plt.style.use(mpl_rc)


def create_hist(theta2_on, theta2_off, cuts: MutableMapping[str, Any]):
    """Create the theta2 histogram assuming a given theta2 global cut."""
    nbins = round((cuts["theta2_range"][1] / cuts["theta2_global_cut"]) * 2)
    hist_on, bin_edges_on = np.histogram(
        theta2_on, density=False, bins=nbins, range=tuple(cuts["theta2_range"])
    )
    hist_off, bin_edges_off = np.histogram(
        theta2_off, density=False, bins=nbins, range=tuple(cuts["theta2_range"])
    )

    bin_width = bin_edges_on[1] - bin_edges_off[0]
    bin_center = bin_edges_on[:-1] + (bin_width / 2)

    return hist_on, hist_off, bin_edges_on, bin_edges_off, bin_center


def lima_significance(
    hist_on,
    hist_off,
    bin_edges_on,
    bin_edges_off,
    eff_time,
    cuts: MutableMapping[str, Any],
):
    """Return the text containing LiMa significance and statistics for plotting them."""
    N_on = np.sum(hist_on[bin_edges_on[1:] <= cuts["theta2_global_cut"]])
    N_off = np.sum(hist_off[bin_edges_off[1:] <= cuts["theta2_global_cut"]])

    idx_min = (np.abs(bin_edges_on - cuts["theta2_norm_range_min"])).argmin()
    idx_max = (np.abs(bin_edges_on - cuts["theta2_norm_range_max"])).argmin()

    Non_norm = np.sum(hist_on[idx_min:idx_max])
    Noff_norm = np.sum(hist_off[idx_min:idx_max])

    alpha = Noff_norm / Non_norm

    stat = WStatCountsStatistic(n_on=N_on, n_off=N_off, alpha=alpha)
    significance_lima = stat.sqrt_ts

    text_statistics = (
        f"N$_{{\\rm on}}$ = {N_on:.0f}\n"
        f"N$_{{\\rm off}}$ = {N_off:.0f}\n"
        f"Livetime = {eff_time.to(u.h):.1f}\n"
        f"Li&Ma significance = {significance_lima:.1f} $\\sigma$"
    )
    log.info(text_statistics)

    box_color = "yellow" if significance_lima > 5 else "white"

    return text_statistics, box_color


def event_selection(data: pd.DataFrame, cuts: MutableMapping[str, Any]) -> pd.DataFrame:
    """Return the dataframe with the selected events."""
    gammaness = np.array(data.gammaness)
    event_type = np.array(data.event_type)

    log.info(
        f'Gammaness global cut: {cuts["gammaness_global_cut"]}\n'
        f'Theta2 global cut: {cuts["theta2_global_cut"]}\n'
    )
    # Mask for event selection
    condition = (
        (gammaness > cuts["gammaness_global_cut"])
        & (event_type != EventType.FLATFIELD.value)
        & (event_type != EventType.SKY_PEDESTAL.value)
    )
    return data[condition]


def plot_theta2(
    bin_center,
    hist_on,
    hist_off,
    legend_text,
    box_color,
    source_name,
    date_obs,
    runs,
    highlevel_dir,
    cuts: MutableMapping[str, Any],
):
    """Plot theta2 histogram and save the figure."""
    _, ax = plt.subplots()

    ax.errorbar(
        bin_center,
        hist_on,
        yerr=np.sqrt(hist_on),
        fmt="o",
        label="ON data",
    )
    ax.errorbar(
        bin_center,
        hist_off,
        yerr=np.sqrt(hist_off),
        fmt="s",
        label="Background",
    )
    ax.set_xlim(0, 0.5)
    ax.axvline(cuts["theta2_global_cut"], color="black", ls="--", alpha=0.75)
    ax.set_xlabel("$\\theta^{2}$ [deg$^{2}$]")
    ax.set_ylabel("Counts")
    ax.legend(title=legend_text, facecolor=box_color, loc="upper right")._legend_box.align = "left"
    ax.set_title(f"Source: {source_name}. Date: {date_obs.strftime('%Y-%m-%d')}\n Runs: {runs}")
    plot_path = highlevel_dir / f"Theta2_{source_name}_{date_obs.strftime('%Y-%m-%d')}.png"
    plt.savefig(plot_path, dpi=300)
    return plot_path


@click.command()
@click.argument("telescope", type=click.Choice(["LST1"]))
@click.option(
    "-d",
    "--date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=YESTERDAY.strftime("%Y-%m-%d"),
)
@click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False),
    default=DEFAULT_CFG,
    help="Read option defaults from the specified cfg file",
)
@click.option("-s", "--simulate", is_flag=True)
def main(
    date: datetime = YESTERDAY,
    telescope: str = "LST1",
    config: Path = DEFAULT_CFG,
    simulate: bool = False,
):
    """Plot theta2 histograms for each source from a given date."""
    log.setLevel(logging.INFO)

    log.debug(f"Config: {config.resolve()}")

    # Initial setup of global parameters
    options.date = date
    flat_date = date_to_dir(date)
    options.tel_id = telescope
    options.prod_id = get_prod_id()
    options.dl2_prod_id = get_dl2_prod_id()
    options.directory = analysis_path(options.tel_id)
    dl2_directory = Path(cfg.get("LST1", "DL2_DIR"))
    highlevel_directory = destination_dir("HIGH_LEVEL", create_dir=True)
    host = cfg.get("WEBSERVER", "HOST")
    cuts = toml.load(SELECTION_CUTS_FILE)

    sources = get_source_list(date)
    log.info(f"Sources: {sources}")

    for source in sources:
        df = pd.DataFrame()
        runs = sources[source]
        log.info(f"Source: {source}, runs: {runs}")

        if simulate:
            continue

        for run in runs:
            input_file = (
                dl2_directory
                / flat_date
                / options.prod_id
                / options.dl2_prod_id
                / f"dl2_LST-1.Run{run:05d}.h5"
            )
            df = pd.concat([df, pd.read_hdf(input_file, key=dl2_params_lstcam_key)])

        selected_events = event_selection(data=df, cuts=cuts)

        try:
            true_source_position = extract_source_position(
                data=selected_events, observed_source_name=source
            )
            off_source_position = [element * -1 for element in true_source_position]

            theta2_on = np.array(compute_theta2(selected_events, true_source_position))
            theta2_off = np.array(compute_theta2(selected_events, off_source_position))

            hist_on, hist_off, bin_edges_on, bin_edges_off, bin_center = create_hist(
                theta2_on, theta2_off, cuts
            )
            text, box_color = lima_significance(
                hist_on=hist_on,
                hist_off=hist_off,
                bin_edges_on=bin_edges_on,
                bin_edges_off=bin_edges_off,
                eff_time=get_effective_time(df)[0],
                cuts=cuts,
            )
            pdf_file = plot_theta2(
                bin_center=bin_center,
                hist_on=hist_on,
                hist_off=hist_off,
                legend_text=text,
                box_color=box_color,
                source_name=source,
                date_obs=date,
                runs=runs,
                highlevel_dir=highlevel_directory,
                cuts=cuts,
            )
            if not simulate:
                dest_directory = directory_in_webserver(
                    host=host, datacheck_type="HIGH_LEVEL", date=flat_date, prod_id=options.prod_id
                )
                cmd = ["scp", pdf_file, f"{host}:{dest_directory}/."]
                subprocess.run(cmd, capture_output=True, check=True)

        except astropy.coordinates.name_resolve.NameResolveError:
            log.warning(f"Source {source} not found in the catalog. Skipping.")
            # TODO: get ra/dec from the TCU database instead


if __name__ == "__main__":
    main()



from datetime import datetime
from pathlib import Path

import pytest

from osa.configs import options
from osa.configs.config import cfg
from osa.utils.utils import date_to_dir


def test_get_raw_dir():
    options.date = datetime.fromisoformat("2020-01-17")
    night_dir = date_to_dir(options.date)
    r0_dir = Path(cfg.get("LST1", "R0_DIR")) / night_dir
    from osa.raw import get_raw_dir

    assert get_raw_dir(options.date) == r0_dir


def test_get_check_raw_dir(r0_dir):
    options.date = datetime.fromisoformat("2020-01-18")
    from osa.raw import get_check_raw_dir

    with pytest.raises(OSError):
        assert get_check_raw_dir(options.date)

    options.date = datetime.fromisoformat("2020-01-17")
    raw_dir = get_check_raw_dir(options.date)
    assert raw_dir.resolve() == r0_dir


def test_is_raw_data_available(r0_data):
    from osa.raw import is_raw_data_available

    for file in r0_data:
        assert file.exists()

    options.date = datetime.fromisoformat("2020-01-17")
    assert is_raw_data_available(options.date) is True



"""Open the lstosa configuration file."""

import configparser
import logging
import sys
from pathlib import Path

from importlib.resources import files

from osa.configs import options
from osa.utils.logging import myLogger

log = myLogger(logging.getLogger(__name__))

__all__ = ["read_config", "cfg", "DEFAULT_CFG"]


DEFAULT_CFG = files("osa").joinpath("configs/sequencer.cfg")


def read_config():
    """
    Read cfg lstosa config file

    Returns
    -------
    config: ConfigParser
        Configuration file cfg
    """

    for idx, arg in enumerate(sys.argv):
        if arg in ["-c", "--config"]:
            options.configfile = sys.argv[idx + 1]
            break
        else:
            options.configfile = DEFAULT_CFG

    file = options.configfile

    if not Path(file).exists():
        raise FileNotFoundError(f"Configuration file {file} not found.")

    config = configparser.ConfigParser(allow_no_value=True)
    try:
        config.read(file)
    except configparser.Error as err:
        log.exception(err)
    log.debug(f"Sections of the config file are {config.sections()}")
    return config


cfg = read_config()



from datetime import datetime

obs_date = datetime.fromisoformat("2020-01-17")


def test_source_list(
    r0_data,
    run_summary_file,
    run_catalog,
    drs4_time_calibration_files,
    systematic_correction_files,
    pedestal_ids_file,
    merged_run_summary,
):
    """Test that the list of name of sources is correct."""
    from osa.nightsummary.extract import get_source_list

    sources = get_source_list(date=obs_date)
    assert list(sources) == ["Crab", "MadeUpSource"]
    assert sources == {"Crab": [1807], "MadeUpSource": [1808]}


def test_build_sequences(sequence_list):
    """Test that building of sequences."""
    from osa.nightsummary.extract import build_sequences

    extracted_seq_list = build_sequences(date=obs_date)
    for sequence, extracted_seq in zip(sequence_list, extracted_seq_list):
        assert sequence.run == extracted_seq.run


def test_no_calib_found(merged_run_summary, run_summary_file_no_calib, run_summary_file_no_calib2):
    """Test that for a day with no PEDCALIB run, the PEDCALIB
     run of another previous day is taken."""
    from osa.nightsummary.extract import extract_runs
    from osa.nightsummary.nightsummary import run_summary_table
    from osa.configs import options

    assert run_summary_file_no_calib.exists()
    assert run_summary_file_no_calib2.exists()
    assert merged_run_summary.exists()

    date = datetime.fromisoformat("2022-09-23")
    options.date = date
    summary_table = run_summary_table(date)
    run_list = extract_runs(summary_table)

    for run in run_list:
        if run.type == 'PEDCALIB':
            pedcalib_run = run

    assert pedcalib_run.run == 9258
    assert pedcalib_run.night == "2022-09-22"

    options.date = datetime.fromisoformat("2020-01-17")



#!/usr/bin/env python

"""Provenance post processing script for OSA pipeline."""

import copy
import logging
import shutil
import sys
from pathlib import Path, PurePath

import yaml

from osa.configs import options
from osa.configs.config import cfg
from osa.provenance.capture import get_activity_id, get_file_hash
from osa.provenance.io import provdoc2graph, provdoc2json, provlist2provdoc, read_prov
from osa.provenance.utils import get_log_config
from osa.utils.cliopts import provprocessparsing
from osa.utils.logging import myLogger

__all__ = ["copy_used_file", "parse_lines_log", "parse_lines_run", "produce_provenance"]

log = myLogger(logging.getLogger())

provconfig = yaml.safe_load(get_log_config())
LOG_FILENAME = provconfig["handlers"]["provHandler"]["filename"]
PROV_PREFIX = provconfig["PREFIX"]
PATH_DL1 = cfg.get("LST1", "DL1_DIR")
PATH_DL2 = cfg.get("LST1", "DL2_DIR")


def copy_used_file(src, outdir):
    """
    Copy file used in process.

    Parameters
    ----------
    src
    outdir
    """
    # check src file exists
    if not Path(src).is_file():
        log.warning(f"{src} file cannot be accessed")

    hash_src = get_file_hash(src, buffer="content")
    filename = PurePath(src).name
    destpath = Path(outdir) / filename
    hash_out = ""

    # get hash and new name
    if destpath.exists():
        hash_out = get_file_hash(str(destpath), buffer="content")
        filename = filename + "_"
        destpath = Path(outdir) / filename

    # try copy file
    if hash_src != hash_out:
        try:
            shutil.copyfile(src, str(destpath))
            log.info(f"copying {destpath}")
        except Exception as ex:
            log.warning(f"could not copy {src} file into {destpath}: {ex}")


def parse_lines_log(filter_cut, calib_runs, run_number):
    """
    Filter content in log file to produce a run/process wise session log.

    Parameters
    ----------
    filter_cut
    calib_runs
    run_number

    Returns
    -------
    filtered
    """
    filtered = []
    if not filter_cut:
        filter_cut = "all"
    cuts = {
        "calibration": ["drs4_pedestal", "calibrate_charge"],
        "r0_to_dl1": ["r0_to_dl1", "dl1ab", "dl1_datacheck"],
        "dl1_to_dl2": ["dl1_to_dl2"],
    }
    cuts["all"] = cuts["calibration"] + cuts["r0_to_dl1"] + cuts["dl1_to_dl2"]

    with open(LOG_FILENAME, "r") as f:
        for line in f.readlines():
            ll = line.split(PROV_PREFIX)
            if len(ll) != 3:
                log.warning(f"format {PROV_PREFIX} mismatch in log file {LOG_FILENAME}\n{line}")
                continue
            prov_str = ll.pop()
            prov_dict = yaml.safe_load(prov_str)
            keep = False
            session_tag = prov_dict.get("session_tag", "0:0")
            session_id = prov_dict.get("session_id", False)
            tag_activity, tag_run = session_tag.split(":")

            # filter by run and calib runs
            if tag_run in [run_number, calib_runs]:
                keep = True
            # filter by activity
            if tag_activity not in cuts[filter_cut]:
                keep = False
            # only keep first session start
            if session_id and (tag_run in [run_number, calib_runs]):
                keep = True
            # make session starts with calibration
            if session_id and filter_cut == "all" and not filtered:
                prov_dict["session_id"] = f"{options.date}{run_number}"
                prov_dict["name"] = run_number
                prov_dict["observation_run"] = run_number
                line = f"{ll[0]}{PROV_PREFIX}{ll[1]}{PROV_PREFIX}{prov_dict}\n"
            # remove parallel sessions
            if session_id and filtered:
                keep = False
            if keep:
                filtered.append(line)

    return filtered


def parse_lines_run(filter_step, prov_lines, out):
    """
    Process provenance info to reduce session at run/process wise scope.

    Parameters
    ----------
    filter_step
    prov_lines
    out

    Returns
    -------
    working_lines
    """
    size = 0
    container = {}
    working_lines = []
    r0filepath_str = ""
    dl1filepath_str = ""
    dl2filepath_str = ""
    mufilepath_str = ""
    ckfilepath_str = ""
    id_activity_run = ""
    end_time_line = ""
    osa_config_copied = False
    for line in prov_lines:
        # get info
        remove = False
        endTime = line.get("endTime", "")
        session_id = line.get("session_id", "")
        activity_id = line.get("activity_id", "")
        filepath = line.get("filepath", "")
        used_role = line.get("used_role", "")
        generated_role = line.get("generated_role", "")
        parameters = line.get("parameters", "")
        name = line.get("name", "")
        content_type = line.get("contentType", "")
        used_id = line.get("used_id", "")
        osa_cfg = line.get("config_file", "")

        # filter grain
        session_tag = line.get("session_tag", "0:0")
        tag_activity, _ = session_tag.split(":")
        if tag_activity != filter_step and not session_id:
            continue

        # remove subruns info
        if name == "DL1CheckSubrunDataset":
            ckfilepath_str = filepath
        elif name == "DL1SubrunDataset":
            dl1filepath_str = filepath
        elif name == "DL2SubrunDataset":
            dl2filepath_str = filepath
        elif name == "MuonsSubrunDataset":
            mufilepath_str = filepath
        elif name == "R0SubrunDataset":
            r0filepath_str = filepath
        if "Subrun" in name or "subrun" in used_role or "subrun" in generated_role:
            remove = True
        if parameters and "ObservationSubRun" in parameters:
            del line["parameters"]["ObservationSubRun"]

        # remove sub-runs activities and info
        if name == filter_step and not id_activity_run:
            id_activity_run = get_activity_id()
        if name in container or used_id in container:
            remove = True
        if parameters and "parameters" in container:
            remove = True
        if name:
            container[name] = True
        if used_id:
            container[used_id] = True
        if parameters:
            container["parameters"] = True
        if endTime:
            remove = True
            end_time_line = line
            size += 1

        # remove duplicated produced files
        if generated_role in container:
            remove = True
        if name == "DL2MergedFile":
            container[name] = True
        if "merged" in generated_role:
            container[generated_role] = True
        if name == "DL1CheckHDF5File":
            container[name] = True
        if "DL1Check HDF5 file" in generated_role:
            container[generated_role] = True
        if name == "DL1CheckPDFFile":
            container[name] = True
        if "DL1Check PDF file" in generated_role:
            container[generated_role] = True

        # replace with new run-wise activity_id
        if activity_id:
            line["activity_id"] = id_activity_run

        # copy used files not subruns not RFs not mergedDL2
        if (
            filepath
            and content_type != "application/x-spss-sav"
            and name != "DL2MergedFile"
            and not name.startswith("DL1Check")
            and not remove
        ):
            copy_used_file(filepath, out)
        if session_id and osa_cfg and not osa_config_copied:
            copy_used_file(osa_cfg, out)
            osa_config_copied = True

        if not remove:
            working_lines.append(line)

    # append collections used and generated at endtime line of last activity
    if end_time_line:
        working_lines.append(end_time_line)
        if r0filepath_str and filter_step == "r0_to_dl1":
            r0_entity_id = get_file_hash(r0filepath_str + "r0", buffer="path")
            r0filepath_str = r0filepath_str.replace(PurePath(r0filepath_str).name, "")
            used = {"entity_id": r0_entity_id}
            used.update({"name": "R0Collection"})
            used.update({"type": "SetCollection"})
            used.update({"size": size})
            used.update({"filepath": r0filepath_str})
            working_lines.append(used)
            used = {"activity_id": id_activity_run}
            used.update({"used_id": r0_entity_id})
            used.update({"used_role": "R0 Collection"})
            working_lines.append(used)
        if dl1filepath_str:
            dl1filepath_str = dl1filepath_str.replace(PurePath(dl1filepath_str).name, "")
            dl1_entity_id = get_file_hash(dl1filepath_str + "dl1", buffer="path")
            dl1 = {"entity_id": dl1_entity_id}
            dl1.update({"name": "DL1Collection"})
            dl1.update({"type": "SetCollection"})
            dl1.update({"size": size})
            dl1.update({"filepath": dl1filepath_str})
            working_lines.append(dl1)
        if mufilepath_str:
            mufilepath_str = mufilepath_str.replace(PurePath(mufilepath_str).name, "")
            mu_entity_id = get_file_hash(mufilepath_str + "muons", buffer="path")
            muons = {"entity_id": mu_entity_id}
            muons.update({"name": "MuonsCollection"})
            muons.update({"type": "SetCollection"})
            muons.update({"size": size})
            muons.update({"filepath": mufilepath_str})
            working_lines.append(muons)
        if mufilepath_str and filter_step == "r0_to_dl1":
            generated = {"activity_id": id_activity_run}
            generated.update({"generated_id": mu_entity_id})
            generated.update({"generated_role": "Muons Collection"})
            working_lines.append(generated)
        if dl1filepath_str and filter_step in ["r0_to_dl1", "dl1ab"]:
            generated = {"activity_id": id_activity_run}
            generated.update({"generated_id": dl1_entity_id})
            generated.update({"generated_role": "DL1 Collection"})
            working_lines.append(generated)
        if dl1filepath_str and filter_step in ["dl1_to_dl2", "dl1ab"]:
            used = {"activity_id": id_activity_run}
            used.update({"used_id": dl1_entity_id})
            used.update({"used_role": "DL1 Collection"})
            working_lines.append(used)
        if dl1filepath_str and filter_step == "dl1_datacheck":
            used = {"activity_id": id_activity_run}
            used.update({"used_id": dl1_entity_id})
            used.update({"used_role": "DL1 Collection"})
            working_lines.append(used)
        if mufilepath_str and filter_step == "dl1_datacheck":
            used = {"activity_id": id_activity_run}
            used.update({"used_id": mu_entity_id})
            used.update({"used_role": "Muons Collection"})
            working_lines.append(used)
        if ckfilepath_str and filter_step == "dl1_datacheck":
            ckfilepath_str = ckfilepath_str.replace(PurePath(ckfilepath_str).name, "")
            chk_entity_id = get_file_hash(ckfilepath_str + "check", buffer="path")
            dl1check = {"entity_id": chk_entity_id}
            dl1check.update({"name": "DL1CheckCollection"})
            dl1check.update({"type": "SetCollection"})
            dl1check.update({"size": size})
            dl1check.update({"filepath": ckfilepath_str})
            working_lines.append(dl1check)
            generated = {"activity_id": id_activity_run}
            generated.update({"generated_id": chk_entity_id})
            generated.update({"generated_role": "DL1Checks Collection"})
            working_lines.append(generated)
        if dl2filepath_str and filter_step == "dl1_to_dl2":
            dl2_entity_id = get_file_hash(dl2filepath_str + "dl2", buffer="path")
            dl2filepath_str = dl2filepath_str.replace(PurePath(dl2filepath_str).name, "")
            used = {"entity_id": dl2_entity_id}
            used.update({"name": "DL2Collection"})
            used.update({"type": "SetCollection"})
            used.update({"size": size})
            used.update({"filepath": dl2filepath_str})
            working_lines.append(used)
            used = {"activity_id": id_activity_run}
            used.update({"generated_id": dl2_entity_id})
            used.update({"generated_role": "DL2 Collection"})
            working_lines.append(used)

    else:
        working_lines = []

    return working_lines


def define_paths(grain, start_path, end_path, base_filename):
    """Define target folders according to granularity."""
    paths = {}

    # check destination folder exists
    step_path = Path(start_path) / options.date / options.prod_id / end_path
    if not step_path.exists():
        log.error(f"Path {step_path} does not exist")

    # make folder log/ if does not exist
    paths["out_path"] = step_path / "log"
    paths["out_path"].mkdir(parents=True, exist_ok=True)

    # define paths for prov products
    paths["log_path"] = paths["out_path"] / f"{grain}_{base_filename}.log"
    paths["json_filepath"] = paths["out_path"] / f"{grain}_{base_filename}.json"
    paths["graph_filepath"] = paths["out_path"] / f"{grain}_{base_filename}.pdf"

    return paths


def produce_provenance_files(processed_lines, paths):
    """Create provenance products as JSON logs and graphs."""
    with open(paths["log_path"], "w") as f:
        for line in processed_lines:
            f.write(f"{line}\n")
    log.info(f"creating {paths['log_path']}")
    provdoc = provlist2provdoc(processed_lines)

    # make json
    try:
        provdoc2json(provdoc, str(paths["json_filepath"]))
        log.info(f"creating {paths['json_filepath']}")
    except Exception as ex:
        log.exception(f"problem while creating json: {ex}")
    # make graph
    try:
        provdoc2graph(provdoc, str(paths["graph_filepath"]), "pdf")
        log.info(f"creating {paths['graph_filepath']}")
    except Exception as ex:
        log.exception(f"problem while creating graph: {ex}")


def produce_provenance(session_log_filename, base_filename):
    """
    Create run-wise provenance products as JSON logs
    and graphs according to granularity.
    """

    if options.filter == "calibration" or not options.filter:
        paths_calibration = define_paths(
            "calibration_to_dl1", PATH_DL1, options.dl1_prod_id, base_filename
        )
        plines_drs4 = parse_lines_run(
            "drs4_pedestal",
            read_prov(filename=session_log_filename),
            str(paths_calibration["out_path"]),
        )
        plines_calib = parse_lines_run(
            "calibrate_charge",
            read_prov(filename=session_log_filename),
            str(paths_calibration["out_path"]),
        )
        calibration_lines = plines_drs4 + plines_calib[1:]

    # TODO
    # create calibration prov files only if filtering
    if options.filter == "calibration":
        pass

    if options.filter == "r0_to_dl1" or not options.filter:
        paths_r0_dl1 = define_paths("r0_to_dl1", PATH_DL1, options.dl1_prod_id, base_filename)
        plines_r0 = parse_lines_run(
            "r0_to_dl1",
            read_prov(filename=session_log_filename),
            str(paths_r0_dl1["out_path"]),
        )
        plines_ab = parse_lines_run(
            "dl1ab",
            read_prov(filename=session_log_filename),
            str(paths_r0_dl1["out_path"]),
        )
        plines_check = parse_lines_run(
            "dl1_datacheck",
            read_prov(filename=session_log_filename),
            str(paths_r0_dl1["out_path"]),
        )
        dl1_lines = plines_r0 + plines_ab[1:] + plines_check[1:]

    # create r0_to_dl1 prov files only if filtering
    if options.filter == "r0_to_dl1":
        produce_provenance_files(plines_r0 + plines_ab[1:] + plines_check[1:], paths_r0_dl1)

    if options.filter == "dl1_to_dl2" or not options.filter:
        if not options.no_dl2:
            paths_dl1_dl2 = define_paths("dl1_to_dl2", PATH_DL2, options.dl2_prod_id, base_filename)
            plines_dl2 = parse_lines_run(
                "dl1_to_dl2",
                read_prov(filename=session_log_filename),
                str(paths_dl1_dl2["out_path"]),
            )
            dl1_dl2_lines = plines_dl2

    # create dl1_to_dl2 prov files only if filtering
    if options.filter == "dl1_to_dl2":
        if not options.no_dl2:
            produce_provenance_files(plines_dl2, paths_dl1_dl2)

    # create calibration_to_dl1 and calibration_to_dl2 prov files
    if not options.filter:
        calibration_to_dl1 = define_paths(
            "calibration_to_dl1", PATH_DL1, options.dl1_prod_id, base_filename
        )
        calibration_to_dl1_lines = calibration_lines + dl1_lines[1:]
        lines_dl1 = copy.deepcopy(calibration_to_dl1_lines)
        produce_provenance_files(lines_dl1, calibration_to_dl1)

        if not options.no_dl2:
            calibration_to_dl2 = define_paths(
                "calibration_to_dl2", PATH_DL2, options.dl2_prod_id, base_filename
            )
            calibration_to_dl2_lines = calibration_to_dl1_lines + dl1_dl2_lines[1:]
            lines_dl2 = copy.deepcopy(calibration_to_dl2_lines)
            produce_provenance_files(lines_dl2, calibration_to_dl2)


def main():
    """Extract the provenance information."""
    provprocessparsing()

    # Logging
    if options.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    # check LOG_FILENAME exists
    if not Path(LOG_FILENAME).exists():
        log.error(f"file {LOG_FILENAME} does not exist")

    # check LOG_FILENAME is not empty
    if not Path(LOG_FILENAME).stat().st_size:
        log.warning(f"file {LOG_FILENAME} is empty")
        sys.exit(1)

    # build base_filename
    base_filename = f"{options.run}_prov"
    session_log_filename = f"{base_filename}.log"

    # parse LOG_FILENAME content for a specific run / process
    calib_runs = f"{options.drs4_pedestal_run_id}-{options.pedcal_run_id}"
    parsed_content = parse_lines_log(options.filter, calib_runs, options.run)

    # create temporal session log file
    with open(session_log_filename, "w") as f:
        for line in parsed_content:
            f.write(line)

    try:
        # create run-wise JSON logs and graphs for each
        produce_provenance(session_log_filename, base_filename)
    finally:
        # remove temporal session log file
        remove_session_log_file = Path(session_log_filename)
        remove_session_log_file.unlink()

    # remove LOG_FILENAME
    if options.quit:
        remove_log_file = Path(LOG_FILENAME)
        remove_log_file.unlink()


if __name__ == "__main__":
    main()



import datetime

import pytest

from osa.configs import options
from osa.configs.config import cfg

options.date = datetime.datetime.fromisoformat("2020-01-17")
options.tel_id = "LST1"
options.prod_id = "v0.1.0"


def test_analysis_path(running_analysis_dir):
    from osa.paths import analysis_path

    assert analysis_path("LST1").resolve() == running_analysis_dir


def test_get_lstchain_version():
    from osa.utils.utils import get_lstchain_version
    from lstchain import __version__

    assert get_lstchain_version() == f"v{__version__}"


def test_get_prod_id():
    from osa.utils.utils import get_prod_id

    prod_id = cfg.get(options.tel_id, "PROD_ID")
    assert get_prod_id() == prod_id


def test_date_to_dir():
    from osa.utils.utils import date_to_dir

    assert date_to_dir(options.date) == "20200117"


def test_time_to_seconds():
    from osa.utils.utils import time_to_seconds

    seconds_with_day = time_to_seconds("2-02:27:15")
    assert seconds_with_day == 2 * 24 * 3600 + 2 * 3600 + 27 * 60 + 15
    seconds = time_to_seconds("02:27:15")
    assert seconds == 2 * 3600 + 27 * 60 + 15
    seconds = time_to_seconds("27:15")
    assert seconds == 27 * 60 + 15
    assert time_to_seconds(None) == 0

    with pytest.raises(ValueError):
        time_to_seconds("12.11.11")


def test_stringify():
    from osa.utils.utils import stringify

    assert stringify(["command", "foo", "--bar"]) == "command foo --bar"


def test_gettag():
    from osa.utils.utils import gettag

    assert gettag() == "test_utils.py(test_gettag)"


def test_night_finished_flag(base_test_dir):
    from osa.utils.utils import night_finished_flag

    assert night_finished_flag() == base_test_dir / "OSA/Closer/20200117/v0.1.0/NightFinished.txt"


def test_create_lock(base_test_dir):
    from osa.utils.utils import create_lock

    lock_path = base_test_dir / "test_lock.closed"
    is_closed = create_lock(lock_path)
    assert is_closed is False
